#
# Copyright (c) 2016 Nordic Semiconductor ASA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.
#
#   3. Neither the name of Nordic Semiconductor ASA nor the names of other
#   contributors to this software may be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   4. This software must only be used in or with a processor manufactured by Nordic
#   Semiconductor ASA, or in or with a processor manufactured by a third party that
#   is used in combination with a processor manufactured by Nordic Semiconductor.
#
#   5. Any software provided in binary or object form under this license must not be
#   reverse engineered, decompiled, modified and/or disassembled.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import binascii
import logging
import struct

# Python imports
import time
from datetime import datetime, timedelta

# Python 3rd party imports
from serial import Serial
from serial.serialutil import SerialException

# from pc_ble_driver_py.exceptions    import NordicSemiException
from ..exceptions import NordicSemiException
from ..lister.device_lister import DeviceLister

# Nordic Semiconductor imports
from .dfu_transport import TRANSPORT_LOGGING_LEVEL, DfuEvent, DfuTransport
from .dfu_trigger import DFUTrigger


class ValidationException(NordicSemiException):
    """ "
    Exception used when validation failed
    """

    pass


logger = logging.getLogger(__name__)


class Slip:
    SLIP_BYTE_END = 0o300
    SLIP_BYTE_ESC = 0o333
    SLIP_BYTE_ESC_END = 0o334
    SLIP_BYTE_ESC_ESC = 0o335

    SLIP_STATE_DECODING = 1
    SLIP_STATE_ESC_RECEIVED = 2
    SLIP_STATE_CLEARING_INVALID_PACKET = 3

    @staticmethod
    def encode(data):
        newData = []
        for elem in data:
            if elem == Slip.SLIP_BYTE_END:
                newData.append(Slip.SLIP_BYTE_ESC)
                newData.append(Slip.SLIP_BYTE_ESC_END)
            elif elem == Slip.SLIP_BYTE_ESC:
                newData.append(Slip.SLIP_BYTE_ESC)
                newData.append(Slip.SLIP_BYTE_ESC_ESC)
            else:
                newData.append(elem)
        newData.append(Slip.SLIP_BYTE_END)
        return newData

    @staticmethod
    def decode_add_byte(c, decoded_data, current_state):
        finished = False
        if current_state == Slip.SLIP_STATE_DECODING:
            if c == Slip.SLIP_BYTE_END:
                finished = True
            elif c == Slip.SLIP_BYTE_ESC:
                current_state = Slip.SLIP_STATE_ESC_RECEIVED
            else:
                decoded_data.append(c)
        elif current_state == Slip.SLIP_STATE_ESC_RECEIVED:
            if c == Slip.SLIP_BYTE_ESC_END:
                decoded_data.append(Slip.SLIP_BYTE_END)
                current_state = Slip.SLIP_STATE_DECODING
            elif c == Slip.SLIP_BYTE_ESC_ESC:
                decoded_data.append(Slip.SLIP_BYTE_ESC)
                current_state = Slip.SLIP_STATE_DECODING
            else:
                current_state = Slip.SLIP_STATE_CLEARING_INVALID_PACKET
        elif current_state == Slip.SLIP_STATE_CLEARING_INVALID_PACKET:
            if c == Slip.SLIP_BYTE_END:
                current_state = Slip.SLIP_STATE_DECODING
                decoded_data = []

        return (finished, current_state, decoded_data)


class DFUAdapter:
    def __init__(self, serial_port):
        self.serial_port = serial_port

    def send_message(self, data):
        packet = Slip.encode(data)
        logger.log(TRANSPORT_LOGGING_LEVEL, "SLIP: --> " + str(data))
        try:
            self.serial_port.write(packet)
        except SerialException as e:
            raise NordicSemiException(
                "Writing to serial port failed: " + str(e) + ". "
                "If MSD is enabled on the target device, try to disable it ref. "
                "https://wiki.segger.com/index.php?title=J-Link-OB_SAM3U"
            )

    def get_message(self):
        current_state = Slip.SLIP_STATE_DECODING
        finished = False
        decoded_data = []

        while finished == False:
            byte = self.serial_port.read(1)
            if byte:
                (byte) = struct.unpack("B", byte)[0]
                (finished, current_state, decoded_data) = Slip.decode_add_byte(
                    byte, decoded_data, current_state
                )
            else:
                current_state = Slip.SLIP_STATE_CLEARING_INVALID_PACKET
                return None

        logger.log(TRANSPORT_LOGGING_LEVEL, "SLIP: <-- " + str(decoded_data))

        return decoded_data


class DfuTransportSerial(DfuTransport):

    DEFAULT_BAUD_RATE = 115200
    DEFAULT_FLOW_CONTROL = True
    DEFAULT_TIMEOUT = 30.0  # Timeout time for board response
    DEFAULT_SERIAL_PORT_TIMEOUT = 1.0  # Timeout time on serial port read
    DEFAULT_PRN = 0
    DEFAULT_DO_PING = True

    OP_CODE = {
        "CreateObject": 0x01,
        "SetPRN": 0x02,
        "CalcChecSum": 0x03,
        "Execute": 0x04,
        "ReadError": 0x05,
        "ReadObject": 0x06,
        "GetSerialMTU": 0x07,
        "WriteObject": 0x08,
        "Ping": 0x09,
        "Response": 0x60,
    }

    def __init__(
        self,
        com_port,
        baud_rate=DEFAULT_BAUD_RATE,
        flow_control=DEFAULT_FLOW_CONTROL,
        timeout=DEFAULT_TIMEOUT,
        prn=DEFAULT_PRN,
        do_ping=DEFAULT_DO_PING,
    ):

        super().__init__()
        self.com_port = com_port
        self.baud_rate = baud_rate
        self.flow_control = 1 if flow_control else 0
        self.timeout = timeout
        self.prn = prn
        self.serial_port = None
        self.dfu_adapter = None
        self.ping_id = 0
        self.do_ping = do_ping

        self.mtu = 0

        """:type: serial.Serial """

    def open(self):
        super().open()
        try:
            self.__ensure_bootloader()
            self.serial_port = Serial(
                port=self.com_port,
                baudrate=self.baud_rate,
                rtscts=self.flow_control,
                timeout=self.DEFAULT_SERIAL_PORT_TIMEOUT,
            )
            self.dfu_adapter = DFUAdapter(self.serial_port)
        except OSError as e:
            raise NordicSemiException(
                "Serial port could not be opened on {0}"
                ". Reason: {1}".format(self.com_port, e.strerror)
            )

        if self.do_ping:
            ping_success = False
            start = datetime.now()
            while (
                datetime.now() - start < timedelta(seconds=self.timeout)
                and ping_success == False
            ):
                if self.__ping() == True:
                    ping_success = True

            if ping_success == False:
                raise NordicSemiException("No ping response after opening COM port")

        self.__set_prn()
        self.__get_mtu()

    def close(self):
        super().close()
        self.serial_port.close()

    def send_init_packet(self, init_packet):
        def try_to_recover():
            if response["offset"] == 0 or response["offset"] > len(init_packet):
                # There is no init packet or present init packet is too long.
                return False

            expected_crc = (
                binascii.crc32(init_packet[: response["offset"]]) & 0xFFFFFFFF
            )

            if expected_crc != response["crc"]:
                # Present init packet is invalid.
                return False

            if len(init_packet) > response["offset"]:
                # Send missing part.
                try:
                    self.__stream_data(
                        data=init_packet[response["offset"] :],
                        crc=expected_crc,
                        offset=response["offset"],
                    )
                except ValidationException:
                    return False

            self.__execute()
            return True

        response = self.__select_command()
        assert len(init_packet) <= response["max_size"], "Init command is too long"

        if try_to_recover():
            return

        try:
            self.__create_command(len(init_packet))
            self.__stream_data(data=init_packet)
            self.__execute()
        except ValidationException:
            raise NordicSemiException("Failed to send init packet")

    def send_firmware(self, firmware):
        def try_to_recover():
            if response["offset"] == 0:
                # Nothing to recover
                return

            expected_crc = binascii.crc32(firmware[: response["offset"]]) & 0xFFFFFFFF
            remainder = response["offset"] % response["max_size"]

            if expected_crc != response["crc"]:
                # Invalid CRC. Remove corrupted data.
                response["offset"] -= (
                    remainder if remainder != 0 else response["max_size"]
                )
                response["crc"] = (
                    binascii.crc32(firmware[: response["offset"]]) & 0xFFFFFFFF
                )
                return

            if (remainder != 0) and (response["offset"] != len(firmware)):
                # Send rest of the page.
                try:
                    to_send = firmware[
                        response["offset"] : response["offset"]
                        + response["max_size"]
                        - remainder
                    ]
                    response["crc"] = self.__stream_data(
                        data=to_send, crc=response["crc"], offset=response["offset"]
                    )
                    response["offset"] += len(to_send)
                except ValidationException:
                    # Remove corrupted data.
                    response["offset"] -= remainder
                    response["crc"] = (
                        binascii.crc32(firmware[: response["offset"]]) & 0xFFFFFFFF
                    )
                    return

            self.__execute()
            self._send_event(
                event_type=DfuEvent.PROGRESS_EVENT, progress=response["offset"]
            )

        response = self.__select_data()
        try_to_recover()
        for i in range(response["offset"], len(firmware), response["max_size"]):
            data = firmware[i : i + response["max_size"]]
            try:
                self.__create_data(len(data))
                response["crc"] = self.__stream_data(
                    data=data, crc=response["crc"], offset=i
                )
                self.__execute()
            except ValidationException:
                raise NordicSemiException("Failed to send firmware")

            self._send_event(event_type=DfuEvent.PROGRESS_EVENT, progress=len(data))

    def __ensure_bootloader(self):
        lister = DeviceLister()

        device = None
        start = datetime.now()
        while not device and datetime.now() - start < timedelta(seconds=self.timeout):
            time.sleep(0.5)
            device = lister.get_device(com=self.com_port)

        if device:
            device_serial_number = device.serial_number

            if not self.__is_device_in_bootloader_mode(device):
                retry_count = 10
                wait_time_ms = 500

                trigger = DFUTrigger()
                try:
                    trigger.enter_bootloader_mode(device)
                    logger.info("Serial: DFU bootloader was triggered")
                except NordicSemiException as err:
                    logger.error(err)

                for checks in range(retry_count):
                    logger.info(
                        "Serial: Waiting {} ms for device to enter bootloader {}/{} time".format(
                            500, checks + 1, retry_count
                        )
                    )

                    time.sleep(wait_time_ms / 1000.0)

                    device = lister.get_device(serial_number=device_serial_number)
                    if self.__is_device_in_bootloader_mode(device):
                        self.com_port = device.get_first_available_com_port()
                        break

                trigger.clean()
            if not self.__is_device_in_bootloader_mode(device):
                logger.info(
                    "Serial: Device is either not in bootloader mode, or using an unsupported bootloader."
                )

    def __is_device_in_bootloader_mode(self, device):
        if not device:
            return False

        #  Return true if nrf bootloader or Jlink interface detected.
        return (
            (
                device.vendor_id.lower() == "1915"
                and device.product_id.lower() == "521f"
            )  # nRF52 SDFU USB
            or (
                device.vendor_id.lower() == "1366"
                and device.product_id.lower() == "0105"
            )  # JLink CDC UART Port
            or (
                device.vendor_id.lower() == "1366"
                and device.product_id.lower() == "1015"
            )
        )  # JLink CDC UART Port (MSD)

    def __set_prn(self):
        logger.debug("Serial: Set Packet Receipt Notification {}".format(self.prn))
        self.dfu_adapter.send_message(
            [DfuTransportSerial.OP_CODE["SetPRN"]] + list(struct.pack("<H", self.prn))
        )
        self.__get_response(DfuTransportSerial.OP_CODE["SetPRN"])

    def __get_mtu(self):
        self.dfu_adapter.send_message([DfuTransportSerial.OP_CODE["GetSerialMTU"]])
        response = self.__get_response(DfuTransportSerial.OP_CODE["GetSerialMTU"])

        self.mtu = struct.unpack("<H", bytearray(response))[0]

    def __ping(self):
        self.ping_id = (self.ping_id + 1) % 256

        self.dfu_adapter.send_message(
            [DfuTransportSerial.OP_CODE["Ping"], self.ping_id]
        )
        resp = (
            self.dfu_adapter.get_message()
        )  # Receive raw response to check return code

        if not resp:
            logger.debug("Serial: No ping response")
            return False

        if resp[0] != DfuTransportSerial.OP_CODE["Response"]:
            logger.debug("Serial: No Response: 0x{:02X}".format(resp[0]))
            return False

        if resp[1] != DfuTransportSerial.OP_CODE["Ping"]:
            logger.debug(
                "Serial: Unexpected Executed OP_CODE.\n"
                + "Expected: 0x{:02X} Received: 0x{:02X}".format(
                    DfuTransportSerial.OP_CODE["Ping"], resp[1]
                )
            )
            return False

        if resp[2] != DfuTransport.RES_CODE["Success"]:
            # Returning an error code is seen as good enough. The bootloader is up and running
            return True
        else:
            if struct.unpack("B", bytearray(resp[3:]))[0] == self.ping_id:
                return True
            else:
                return False

    def __create_command(self, size):
        self.__create_object(0x01, size)

    def __create_data(self, size):
        self.__create_object(0x02, size)

    def __create_object(self, object_type, size):
        self.dfu_adapter.send_message(
            [DfuTransportSerial.OP_CODE["CreateObject"], object_type]
            + list(struct.pack("<L", size))
        )
        self.__get_response(DfuTransportSerial.OP_CODE["CreateObject"])

    def __calculate_checksum(self):
        self.dfu_adapter.send_message([DfuTransportSerial.OP_CODE["CalcChecSum"]])
        response = self.__get_response(DfuTransportSerial.OP_CODE["CalcChecSum"])

        if response is None:
            raise NordicSemiException(
                "Did not receive checksum response from DFU target. "
                "If MSD is enabled on the target device, try to disable it ref. "
                "https://wiki.segger.com/index.php?title=J-Link-OB_SAM3U"
            )

        (offset, crc) = struct.unpack("<II", bytearray(response))
        return {"offset": offset, "crc": crc}

    def __execute(self):
        self.dfu_adapter.send_message([DfuTransportSerial.OP_CODE["Execute"]])
        self.__get_response(DfuTransportSerial.OP_CODE["Execute"])

    def __select_command(self):
        return self.__select_object(0x01)

    def __select_data(self):
        return self.__select_object(0x02)

    def __select_object(self, object_type):
        logger.debug("Serial: Selecting Object: type:{}".format(object_type))
        self.dfu_adapter.send_message(
            [DfuTransportSerial.OP_CODE["ReadObject"], object_type]
        )

        response = self.__get_response(DfuTransportSerial.OP_CODE["ReadObject"])
        (max_size, offset, crc) = struct.unpack("<III", bytearray(response))

        logger.debug(
            "Serial: Object selected: "
            + " max_size:{} offset:{} crc:{}".format(max_size, offset, crc)
        )
        return {"max_size": max_size, "offset": offset, "crc": crc}

    def __get_checksum_response(self):
        resp = self.__get_response(DfuTransportSerial.OP_CODE["CalcChecSum"])

        (offset, crc) = struct.unpack("<II", bytearray(resp))
        return {"offset": offset, "crc": crc}

    def __stream_data(self, data, crc=0, offset=0):
        logger.debug(
            "Serial: Streaming Data: "
            + "len:{0} offset:{1} crc:0x{2:08X}".format(len(data), offset, crc)
        )

        def validate_crc():
            if crc != response["crc"]:
                raise ValidationException(
                    "Failed CRC validation.\n"
                    + "Expected: {} Received: {}.".format(crc, response["crc"])
                )
            if offset != response["offset"]:
                raise ValidationException(
                    "Failed offset validation.\n"
                    + "Expected: {} Received: {}.".format(offset, response["offset"])
                )

        current_pnr = 0

        for i in range(0, len(data), (self.mtu - 1) // 2 - 1):
            # append the write data opcode to the front
            # here the maximum data size is self.mtu/2,
            # due to the slip encoding which at maximum doubles the size
            to_transmit = data[i : i + (self.mtu - 1) // 2 - 1]
            to_transmit = (
                struct.pack("B", DfuTransportSerial.OP_CODE["WriteObject"])
                + to_transmit
            )

            self.dfu_adapter.send_message(list(to_transmit))
            crc = binascii.crc32(to_transmit[1:], crc) & 0xFFFFFFFF
            offset += len(to_transmit) - 1
            current_pnr += 1
            if self.prn == current_pnr:
                current_pnr = 0
                response = self.__get_checksum_response()
                validate_crc()
        response = self.__calculate_checksum()
        validate_crc()
        return crc

    def __get_response(self, operation):
        def get_dict_key(dictionary, value):
            return next(
                (key for key, val in list(dictionary.items()) if val == value), None
            )

        resp = self.dfu_adapter.get_message()

        if not resp:
            return None

        if resp[0] != DfuTransportSerial.OP_CODE["Response"]:
            raise NordicSemiException("No Response: 0x{:02X}".format(resp[0]))

        if resp[1] != operation:
            raise NordicSemiException(
                "Unexpected Executed OP_CODE.\n"
                + "Expected: 0x{:02X} Received: 0x{:02X}".format(operation, resp[1])
            )

        if resp[2] == DfuTransport.RES_CODE["Success"]:
            return resp[3:]

        elif resp[2] == DfuTransport.RES_CODE["ExtendedError"]:
            try:
                data = DfuTransport.EXT_ERROR_CODE[resp[3]]
            except IndexError:
                data = "Unsupported extended error type {}".format(resp[3])
            raise NordicSemiException(
                "Extended Error 0x{:02X}: {}".format(resp[3], data)
            )
        else:
            raise NordicSemiException(
                "Response Code {}".format(get_dict_key(DfuTransport.RES_CODE, resp[2]))
            )



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Copyright 2020 Nitrokey Gmbh

Licensed under the Apache License, Version 2.0, <LICENSE-APACHE or
http://apache.org/licenses/LICENSE-2.0> or the MIT license <LICENSE-MIT or
http://opensource.org/licenses/MIT>, at your option. This file may not be
copied, modified, or distributed except according to those terms.
"""

import logging
import subprocess
import threading
import time
from sys import stderr
from typing import List, Optional


class ThreadLog(threading.Thread):
    _dmesg_skip_strings = []  # type: ignore

    _write_to_log = False

    def __init__(self, logger: logging.Logger, command: str):
        threading.Thread.__init__(self)
        self.finished = False
        self.logger = logger
        self.command = command
        self.daemon = True
        self.start()
        self.process: Optional[subprocess.Popen[bytes]] = None

    def run(self):
        self.execute(self.command.split())

    @staticmethod
    def _contains(value, strings):
        for s in strings:
            if s in value:
                return True
        return False

    def execute(self, command: List[str]):
        self.process = subprocess.Popen(
            command, shell=False, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        )

        # Poll process for new output until finished
        for line in iter(self.process.stdout.readline, ""):  # type: ignore
            if self.finished:
                break
            if not line or self.finished or not self._write_to_log:
                continue
            if self._contains(line, self._dmesg_skip_strings):
                continue
            self.logger.debug(line.strip())

        self.process.wait()
        self.logger.debug("Finished")

    def start_logging(self):
        self._write_to_log = True

    def __enter__(self):
        time.sleep(1)
        self.start_logging()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.logger.debug("Finishing")
        self.finished = True
        self.process.kill()
        self.process.wait()
        self.join(10)


def test_run():
    FORMAT = "%(relativeCreated)05d [%(process)x] - %(levelname)s - %(name)s - %(message)s [%(filename)s:%(lineno)d]"
    logging.basicConfig(format=FORMAT, stream=stderr, level=logging.DEBUG)
    logger = logging.getLogger("threadlog")

    try:
        t = ThreadLog(logger, "dmesg -w")
        t.start_logging()
        time.sleep(10)
        t.finished = True
    except (KeyboardInterrupt, SystemExit) as k:
        return


if __name__ == "__main__":
    test_run()



import enum
import sys
from dataclasses import dataclass
from enum import Enum, IntFlag
from typing import Optional

from fido2 import cbor
from fido2.ctap import CtapError

from pynitrokey.helpers import local_critical, local_print
from pynitrokey.nk3.device import Command, Nitrokey3Device

from .device import VERSION_LEN
from .utils import Version


@enum.unique
class AdminCommand(Enum):
    STATUS = 0x80
    TEST_SE050 = 0x81
    GET_CONFIG = 0x82
    SET_CONFIG = 0x83
    FACTORY_RESET = 0x84
    FACTORY_RESET_APP = 0x85


@enum.unique
class InitStatus(IntFlag):
    NFC_ERROR = 0b0001
    INTERNAL_FLASH_ERROR = 0b0010
    EXTERNAL_FLASH_ERROR = 0b0100
    MIGRATION_ERROR = 0b1000
    SE050_ERROR = 0b00010000

    def is_error(self) -> bool:
        return self.value != 0

    def __str__(self) -> str:
        if self == 0:
            return "ok"
        errors = [error for error in InitStatus if error in self if error.name]
        value = sum(errors)
        messages = [error.name for error in errors if error.name]
        if self.value != value:
            messages.append("UNKNOWN")
        return ", ".join(messages) + " (" + hex(self.value) + ")"


@enum.unique
class Variant(Enum):
    USBIP = 0
    LPC55 = 1
    NRF52 = 2


@dataclass
class Status:
    init_status: Optional[InitStatus] = None
    ifs_blocks: Optional[int] = None
    efs_blocks: Optional[int] = None
    variant: Optional[Variant] = None


@enum.unique
class FactoryResetStatus(Enum):
    SUCCESS = 0
    NOT_CONFIRMED = 0x01
    APP_NOT_ALLOWED = 0x02
    APP_FAILED_PARSE = 0x03

    @classmethod
    def from_int(cls, i: int) -> Optional["FactoryResetStatus"]:
        for status in FactoryResetStatus:
            if status.value == i:
                return status
        return None

    @classmethod
    def check(cls, i: int, msg: str) -> None:
        status = FactoryResetStatus.from_int(i)
        if status != FactoryResetStatus.SUCCESS:
            if status is None:
                raise Exception(f"Unknown error {i:x}")
            if status == FactoryResetStatus.NOT_CONFIRMED:
                error = "Operation was not confirmed with touch"
            elif status == FactoryResetStatus.APP_NOT_ALLOWED:
                error = "The application does not support factory reset through nitropy"
            elif status == FactoryResetStatus.APP_FAILED_PARSE:
                error = "The application name must be utf-8"
            local_critical(f"{msg}: {error}", support_hint=False)


@enum.unique
class ConfigStatus(Enum):
    SUCCESS = 0
    READ_FAILED = 1
    WRITE_FAILED = 2
    DESERIALIZATION_FAILED = 3
    SERIALIZATION_FAILED = 4
    INVALID_KEY = 5
    INVALID_VALUE = 6
    DATA_TOO_LONG = 7

    @classmethod
    def from_int(cls, i: int) -> Optional["ConfigStatus"]:
        for status in ConfigStatus:
            if status.value == i:
                return status
        return None

    @classmethod
    def check(cls, i: int, msg: str) -> None:
        status = ConfigStatus.from_int(i)
        if status != ConfigStatus.SUCCESS:
            if status:
                error = str(status)
            else:
                error = f"unknown error {i:x}"
            raise Exception(f"{msg}: {error}")


class AdminApp:
    def __init__(self, device: Nitrokey3Device) -> None:
        self.device = device

    def _call(
        self,
        command: AdminCommand,
        response_len: Optional[int] = None,
        data: bytes = b"",
    ) -> Optional[bytes]:
        try:
            return self.device._call(
                Command.ADMIN,
                response_len=response_len,
                data=command.value.to_bytes(1, "big") + data,
            )
        except CtapError as e:
            if e.code == CtapError.ERR.INVALID_COMMAND:
                return None
            else:
                raise

    def status(self) -> Status:
        status = Status()
        reply = self._call(AdminCommand.STATUS)
        if reply is not None:
            if not reply:
                raise ValueError("The device returned an empty status")
            status.init_status = InitStatus(reply[0])
            if len(reply) >= 4:
                status.ifs_blocks = reply[1]
                status.efs_blocks = int.from_bytes(reply[2:4], "big")
            if len(reply) >= 5:
                try:
                    status.variant = Variant(reply[4])
                except ValueError:
                    pass
        return status

    def version(self) -> Version:
        reply = self.device._call(Command.VERSION, data=bytes([0x01]))
        if len(reply) == VERSION_LEN:
            version = int.from_bytes(reply, "big")
            return Version.from_int(version)
        else:
            return Version.from_str(reply.decode("utf-8"))

    def se050_tests(self) -> Optional[bytes]:
        return self._call(AdminCommand.TEST_SE050)

    def has_config(self, key: str) -> bool:
        reply = self._call(AdminCommand.GET_CONFIG, data=key.encode())
        if not reply or len(reply) < 1:
            return False
        return ConfigStatus.from_int(reply[0]) == ConfigStatus.SUCCESS

    def get_config(self, key: str) -> str:
        reply = self._call(AdminCommand.GET_CONFIG, data=key.encode())
        if not reply or len(reply) < 1:
            raise ValueError("The device returned an empty response")
        ConfigStatus.check(reply[0], "Failed to get config value")
        return reply[1:].decode()

    def set_config(self, key: str, value: str) -> None:
        request = cbor.encode({"key": key, "value": value})
        reply = self._call(AdminCommand.SET_CONFIG, data=request, response_len=1)
        assert reply
        ConfigStatus.check(reply[0], "Failed to set config value")

    def factory_reset(self) -> None:
        try:
            local_print(
                "Please touch the device to confirm the operation", file=sys.stderr
            )
            reply = self._call(AdminCommand.FACTORY_RESET, response_len=1)
            if reply is None:
                local_critical(
                    "Factory reset is not supported by the firmware version on the device",
                    support_hint=False,
                )
                return
        except OSError as e:
            if e.errno == 5:
                self.device.logger.debug("ignoring OSError after reboot", exc_info=e)
                return
            else:
                raise e
        FactoryResetStatus.check(reply[0], "Failed to factory reset the device")

    def factory_reset_app(self, application: str) -> None:
        local_print("Please touch the device to confirm the operation", file=sys.stderr)
        reply = self._call(
            AdminCommand.FACTORY_RESET_APP,
            data=application.encode("ascii"),
            response_len=1,
        )
        if reply is None:
            local_critical(
                "Application Factory reset is not supported by the firmware version on the device",
                support_hint=False,
            )
            return
        FactoryResetStatus.check(reply[0], "Failed to factory reset the device")



"""
kdf_calc.py - a library for calculating hash by KDF

Copyright (C) 2018  Free Software Initiative of Japan
Author: NIIBE Yutaka <gniibe@fsij.org>

This file is a part of Gnuk, a GnuPG USB Token implementation.

Gnuk is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Gnuk is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from cffi import FFI

DEF_gcry_kdf_derive = """
typedef unsigned int gpg_error_t;
gpg_error_t gcry_kdf_derive (const void *passphrase, size_t passphraselen,
                             int algo, int subalgo, const void *salt,
                             size_t saltlen, unsigned long iterations,
                             size_t keysize, void *keybuffer);
"""

GCRY_KDF_ITERSALTED_S2K = 19
GCRY_MD_SHA256 = 8


def kdf_calc(pw_string, salt_byte, iterations):
    ffi = FFI()
    ffi.cdef(DEF_gcry_kdf_derive)
    libgcrypt = ffi.dlopen("libgcrypt.so.20")
    if isinstance(pw_string, str):
        pw_byte = pw_string.encode("UTF-8")
    else:
        pw_byte = pw_string
    pw = ffi.new("char []", pw_byte)
    salt = ffi.new("char []", salt_byte)
    kb = ffi.new("char []", 32)
    r = libgcrypt.gcry_kdf_derive(
        pw,
        len(pw_string),
        GCRY_KDF_ITERSALTED_S2K,
        GCRY_MD_SHA256,
        salt,
        8,
        iterations,
        32,
        kb,
    )
    if r != 0:
        raise ValueError("libgcrypt error", r)
    return ffi.unpack(kb, 32)



#
# Copyright (c) 2016 Nordic Semiconductor ASA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.
#
#   3. Neither the name of Nordic Semiconductor ASA nor the names of other
#   contributors to this software may be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   4. This software must only be used in or with a processor manufactured by Nordic
#   Semiconductor ASA, or in or with a processor manufactured by a third party that
#   is used in combination with a processor manufactured by Nordic Semiconductor.
#
#   5. Any software provided in binary or object form under this license must not be
#   reverse engineered, decompiled, modified and/or disassembled.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from enum import Enum


class HexType(Enum):
    SOFTDEVICE = 1
    BOOTLOADER = 2
    SD_BL = 3
    APPLICATION = 4
    EXTERNAL_APPLICATION = 5


class FirmwareKeys(Enum):
    ENCRYPT = 1
    FIRMWARE_FILENAME = 2
    BIN_FILENAME = 3
    DAT_FILENAME = 4
    INIT_PACKET_DATA = 5
    SD_SIZE = 6
    BL_SIZE = 7
    BOOT_VALIDATION_TYPE = 8



# -*- coding: utf-8 -*-
#
# Copyright 2019 SoloKeys Developers
#
# Licensed under the Apache License, Version 2.0, <LICENSE-APACHE or
# http://apache.org/licenses/LICENSE-2.0> or the MIT license <LICENSE-MIT or
# http://opensource.org/licenses/MIT>, at your option. This file may not be
# copied, modified, or distributed except according to those terms.


import sys
import time

import click
import serial


@click.command()
@click.argument("serial_port")
def monitor(serial_port: str) -> None:
    """Reads Nitrokey  serial output from USB serial port SERIAL_PORT.

    SERIAL-PORT is something like /dev/ttyACM0 or COM10.
    Automatically reconnects. Baud rate is 115200.
    """

    ser = None
    while True:
        try:
            ser = serial.Serial(serial_port, 115200, timeout=0.05)
            break
        except KeyboardInterrupt:
            exit(1)
        except:
            sys.stdout.buffer.write(b".")
            sys.stdout.flush()
        time.sleep(0.5)

    def reconnect() -> serial.Serial:
        while True:
            time.sleep(0.5)
            try:
                ser = serial.Serial(serial_port, 115200, timeout=0.05)
                return ser
            except serial.SerialException:
                sys.stdout.buffer.write(b".")
                sys.stdout.flush()
                pass

    t0 = time.time()
    while True:
        try:
            data = ser.read(1)
            if b"\n" in data:
                t1 = time.time()
                times = "{0:5.3f}".format(t1 - t0)
                data = f"{data.decode()} {times} ".encode()
            sys.stdout.buffer.write(data)
        except KeyboardInterrupt:
            print("\nClosing")
            ser.close()
            return
        except serial.SerialException:
            sys.stdout.buffer.write(b".\n")
            print("reconnecting...")
            # ser = reconnect()
            # ser.close()
            time.sleep(1.0)
            ser = reconnect()
            print("done")
        sys.stdout.flush()



# -*- coding: utf-8 -*-
#
# Copyright 2019 SoloKeys Developers
#
# Licensed under the Apache License, Version 2.0, <LICENSE-APACHE or
# http://apache.org/licenses/LICENSE-2.0> or the MIT license <LICENSE-MIT or
# http://opensource.org/licenses/MIT>, at your option. This file may not be
# copied, modified, or distributed except according to those terms.

import ctypes
import functools
import logging
import os
import platform
import sys
import time
from getpass import getpass
from importlib.metadata import version
from itertools import chain
from numbers import Number
from threading import Event, Timer
from typing import Any, Callable, Dict, List, NoReturn, Optional, Tuple, TypeVar, Union

import click
from semver.version import Version
from tqdm import tqdm

from pynitrokey.confconsts import (
    CLI_LOG_BLACKLIST,
    LOG_FN,
    SUPPORT_EMAIL,
    SUPPORT_URL,
    UDEV_URL,
    VERBOSE,
    Verbosity,
)
from pynitrokey.updates import Repository

STDOUT_PRINT = True


def normalize_parameters(s: str) -> list[str]:
    """Helper function to normalize different writing of parameters

    `s`: `str` parameter to normalize

    Returns:
        `list of str`: normalized form of `parameter`
    """
    if s.startswith("--"):
        return s.split("=", maxsplit=1)
    else:
        return [s]


def filter_sensitive_parameters(parameters: list[str]) -> list[str]:
    """Helper function to remove sensitive parameters

    `parameters`: `list of str`

    Returns:
        `list of str`: `parameters` without sensitive values listed
                       in `pynitrokey.confconsts.CLI_LOG_BLACKLIST`
    """
    parameters = list(
        chain.from_iterable(
            [normalize_parameters(parameter) for parameter in parameters]
        )
    )

    redact_count = 0
    for i in range(len(parameters)):
        if redact_count > 0:
            parameters[i] = "[redacted]"
            redact_count -= 1
        elif parameters[i] in CLI_LOG_BLACKLIST:
            redact_count = CLI_LOG_BLACKLIST[parameters[i]]
    return parameters


def to_websafe(data: str) -> str:
    data = data.replace("+", "-")
    data = data.replace("/", "_")
    data = data.replace("=", "")
    return data


def from_websafe(data: str) -> str:
    data = data.replace("-", "+")
    data = data.replace("_", "/")
    return data + "=="[: (3 * len(data)) % 4]


class ProgressBar:
    """
    Helper class for progress bars where the total length of the progress bar
    is not available before the first iteration.
    """

    def __init__(self, **kwargs: Any) -> None:
        self.bar: Optional[tqdm[Any]] = None
        self.kwargs = kwargs
        self.sum = 0

    def __enter__(self) -> "ProgressBar":
        return self

    def __exit__(self, exc_type: None, exc_val: None, exc_tb: None) -> None:
        self.close()

    def update(self, n: int, total: int) -> None:
        if not self.bar:
            self.bar = tqdm(total=total, **self.kwargs)
        self.bar.update(n)
        self.sum += n

    def update_sum(self, n: int, total: int) -> None:
        if not self.bar:
            self.bar = tqdm(total=total, **self.kwargs)
        if n > self.sum:
            self.bar.update(n - self.sum)
            self.sum = n

    def close(self) -> None:
        if self.bar:
            self.bar.close()


class DownloadProgressBar(ProgressBar):
    """
    Helper class for progress bars for downloading a file.
    """

    def __init__(self, desc: str) -> None:
        super().__init__(desc=f"Download {desc}", unit="B", unit_scale=True)


class Timeout(object):
    """
    Utility class for adding a timeout to an event.
    :param time_or_event: A number, in seconds, or a threading.Event object.
    :ivar event: The Event associated with the Timeout.
    :ivar timer: The Timer associated with the Timeout, if any.
    """

    def __init__(self, time_or_event: Union[float, Event]) -> None:
        if isinstance(time_or_event, float):
            self.event = Event()
            self.timer: Optional[Timer] = Timer(float(time_or_event), self.event.set)
        else:
            self.event = time_or_event
            self.timer = None

    def __enter__(self) -> Event:
        if self.timer:
            self.timer.start()
        return self.event

    def __exit__(self, exc_type: None, exc_val: None, exc_tb: None) -> None:
        if self.timer:
            self.timer.join()
            self.timer.cancel()


class Try:
    """Utility class for an execution of a repeated action with Retries."""

    def __init__(self, i: int, retries: int) -> None:
        self.i = i
        self.retries = retries

    def __str__(self) -> str:
        return f"try {self.i + 1} of {self.retries}"

    def __repr__(self) -> str:
        return f"Try(i={self.i}, retries={self.retries})"


class Retries:
    """Utility class for repeating an action multiple times until it succeeds."""

    def __init__(self, retries: int, timeout: float = 0.5) -> None:
        self.retries = retries
        self.i = 0
        self.timeout = timeout

    def __iter__(self) -> "Retries":
        return self

    def __next__(self) -> Try:
        if self.i >= self.retries:
            raise StopIteration
        if self.i > 0:
            time.sleep(self.timeout)
        t = Try(self.i, self.retries)
        self.i += 1
        return t


# @todo: introduce granularization: dbg, info, err (warn?)
#        + machine-readable
#        + logfile-only (partly solved)
def local_print(*messages: Any, **kwargs: Any) -> None:
    """Application-wide logging function"""

    passed_exc = None
    logger = logging.getLogger()

    for item in messages:
        # handle exception in order as, if it is a regular message
        if isinstance(item, Exception):
            logger.exception(item)
            passed_exc = item
            item = repr(item)
            item = "\tException encountered: " + item

        # just a newline, don't log to file...
        elif item is None or item == "":
            item = ""

        # logfile debug output
        else:
            whereto = "print: " if STDOUT_PRINT else ""
            logger.debug(f"{whereto}{str(item).strip()}")

        # to stdout
        if STDOUT_PRINT:
            print(item, **kwargs)

    # handle `passed_exc`: re-raise on debug verbosity!
    if VERBOSE == Verbosity.debug and passed_exc:
        raise passed_exc


def local_critical(
    *messages: Any, support_hint: bool = True, ret_code: int = 1, **kwargs: Any
) -> None:

    global STDOUT_PRINT
    messages = ("Critical error:",) + tuple(messages)
    local_print(*messages, **kwargs)

    if support_hint:

        # list all connected devices to logfile
        # @fixme: not the best solution
        STDOUT_PRINT = False
        local_print("listing all connected devices:")

        try:
            from pynitrokey.cli import nitropy

            nitropy.commands["list"].callback()  # type: ignore

        except Exception:
            local_print("Unable to list devices. See log for the details.")
            logger = logging.getLogger()
            logger.exception("Unable to list devices")

        STDOUT_PRINT = True

        linux = "linux" in platform.platform().lower()
        local_print(
            "",
            "-" * 80,
            "Critical error occurred, exiting now",
            "Unexpected? Is this a bug? Would you like to get support/help?",
            f"- You can report issues at: {SUPPORT_URL}",
            f"- Writing an e-mail to {SUPPORT_EMAIL} is also possible",
            f"- Please attach the log: '{LOG_FN}' with any support/help request!",
            f"- Please check if you have udev rules installed: {UDEV_URL}"
            if linux
            else "" "-" * 80,
            "",
        )
    sys.exit(ret_code)


# @fixme: consider using/wrapping click.confirm() instead of this...
class AskUser:
    """
    Asking user for input:
        `question`:       printed user question
        `options`:        `[]`        -> we want some data input
                          `List[str]` -> only allow items inside iterable
        `strict`:         if `options` are used, force full match
        `repeat`:         ask `question` up to `repeat` times, if `options` are provided
        `adapt_question`: adapt user-provided `question` (add options, whitespace...),
                          set to `False`, if strictly `question` shall be used
        `hide_input`:     use 'getpass' instead of regular `input`
    """

    def __init__(
        self,
        question: str,
        options: List[str] = [],
        strict: bool = False,
        repeat: int = 3,
        adapt_question: bool = True,
        hide_input: bool = False,
        envvar: Optional[str] = None,
    ) -> None:

        self.data: Optional[str] = None

        self.question = question
        self.adapt_question = adapt_question
        self.final_question = question
        if self.adapt_question:
            _q = self.final_question
            # strip ending colon(s) ':' or whitespace(s) ' '
            _q = _q.strip(" ").strip(":").strip(" ").strip(":")
            if options:
                _q += (
                    f" [{'/'.join(options)}]"
                    if strict
                    else f" [{'/'.join(f'({o[0]}){o[1:]}' for o in options)}]"
                )
            _q += ": "
            self.final_question = _q

        self.options = options
        self.strict = strict
        self.repeat = repeat or 1
        self.hide_input = hide_input
        self.envvar = envvar

    @classmethod
    def yes_no(cls, what: str, strict: bool = False) -> bool:
        opts = ["yes", "no"]
        return cls(what, options=opts, strict=strict).ask() == opts[0]

    @classmethod
    def strict_yes_no(cls, what: str) -> bool:
        return cls.yes_no(what, strict=True)

    @classmethod
    def plain(cls, what: str) -> str:
        return cls(what).ask()

    @classmethod
    def hidden(cls, what: str) -> str:
        return cls(what, hide_input=True).ask()

    def get_input(
        self,
        pre_str: Optional[str] = None,
        hide_input: Optional[Union[str, bool]] = None,
    ) -> str:
        pre_input_string = pre_str or self.final_question
        hide_input = hide_input if hide_input is not None else self.hide_input
        if hide_input:
            return getpass(pre_input_string)
        else:
            print(pre_input_string, end="", file=sys.stderr)
            return input().strip()

    def ask(self) -> str:
        answer = None
        if self.envvar is not None:
            fromvar = os.environ.get(self.envvar)
            if fromvar is not None:
                answer = fromvar

        if answer is None:
            answer = self.get_input()

        # handle plain input request first
        if not self.options:
            self.data = answer
            return self.data

        # now `options` based
        retries = self.repeat
        while retries:
            if answer in self.options:
                self.data = answer
                return self.data

            if not self.strict:
                short_opts = {c[0].lower(): c for c in self.options}
                if len(answer) > 0:
                    self.data = short_opts.get(answer[0].lower())

                if self.data:
                    local_print(f"choosing: {self.data}")
                    return self.data

            answer = self.get_input()
            retries -= 1

        if retries == 0:
            local_critical("max tries exceeded - exiting...")

        assert self.data is None, "expecting `self.data` to be None at this point!"
        return self.data or ""


confirm = functools.partial(click.confirm, err=True)
prompt = functools.partial(click.prompt, err=True)


def require_windows_admin() -> None:
    if os.name == "nt":
        if ctypes.windll.shell32.IsUserAnAdmin() == 0:  # type: ignore
            local_print(
                "Warning: It is recommended to execute nitropy with admin privileges "
                "to be able to access Nitrokey 3 and Nitrokey FIDO 2 devices."
            )


def check_pynitrokey_version() -> None:
    """Checks wether the used pynitrokey version is the latest available version and warns the user if the used version is outdated"""

    latest_release = Repository("Nitrokey", "pynitrokey").get_latest_release()
    latest_version = Version.parse(latest_release.tag[1:])

    current_version = Version.parse(version("pynitrokey"))

    if current_version < latest_version:
        local_print(
            f"You are using an outdated version ({current_version}) of pynitrokey."
        )
        local_print(f"Latest pynitrokey version is {latest_version}")
        local_print("Updating with an outdated version is discouraged.")

        if not confirm("Do you still want to continue?", default=False):
            raise click.Abort()


def check_experimental_flag(experimental: bool) -> None:
    """Helper function to show common warning for the experimental features"""
    if not experimental:
        local_print(" ")
        local_print(
            "This feature is experimental, which means it was not tested thoroughly.\n"
            "Note: data stored with it can be lost in the next firmware update.\n"
            "Please pass --experimental switch to force running it anyway."
        )
        local_print(" ")
        raise click.Abort()



"""
Tests for the Secrets application interface placed in secrets_app.py.
Requires a live device, or an USB-IP simulation.
"""

import binascii
import datetime
import hashlib
import hmac
import logging
import time
from contextlib import suppress
from datetime import timedelta
from os import environ, wait
from sys import stderr
from typing import Any, Callable, List, Optional, Tuple

import fido2
import pytest
import tlv8

from pynitrokey.conftest import (
    CALCULATE_ALL_COMMANDS,
    CHALLENGE,
    CHALLENGE_RESPONSE_COMMANDS,
    CREDID,
    CREDID2,
    DELAY_AFTER_FAILED_REQUEST_SECONDS,
    DIGITS,
    FEATURE_BRUTEFORCE_PROTECTION_ENABLED,
    FEATURE_CHALLENGE_RESPONSE_ENABLED,
    HOTP_WINDOW_SIZE,
    PIN,
    PIN2,
    PIN_ATTEMPT_COUNTER_DEFAULT,
    SECRET,
    CredEncryptionType,
)
from pynitrokey.nk3.secrets_app import (
    Algorithm,
    CCIDInstruction,
    Instruction,
    Kind,
    PasswordSafeEntry,
    RawBytes,
    SecretsApp,
    SecretsAppException,
    Tag,
)

CREDENTIAL_LABEL_MAX_SIZE = 127


def test_reset(secretsAppResetLogin):
    """
    Clear credentials' storage. Simple test.
    """
    secretsAppResetLogin.reset()


def test_list(secretsAppResetLogin):
    """
    List saved credentials. Simple test.
    """
    secretsAppResetLogin.list()


def test_register(secretsAppResetLogin):
    """
    Register credential with the given id and properties. Simple test.
    """
    secretsAppResetLogin.register(CREDID, SECRET, DIGITS, kind=Kind.Hotp)


def test_calculate(secretsAppResetLogin):
    """
    Run calculation on the default credential id. Simple test.
    """
    secretsAppResetLogin.register(CREDID, SECRET, DIGITS, kind=Kind.Hotp)
    secretsAppResetLogin.verify_pin_raw(PIN)
    code = secretsAppResetLogin.calculate(CREDID, CHALLENGE)
    print(code)


def test_delete(secretsAppResetLogin):
    """
    Remove credential with the given id. Simple test.
    """
    secretsAppResetLogin.register(CREDID, SECRET, DIGITS, kind=Kind.Hotp)
    secretsAppResetLogin.verify_pin_raw(PIN)
    secretsAppResetLogin.delete(CREDID)


def test_delete_nonexisting(secretsAppResetLogin):
    """
    Should fail when trying to remove non-existing credential id.
    """
    with pytest.raises(SecretsAppException, match="NotFound"):
        secretsAppResetLogin.delete(CREDID)


def test_list_changes(secretsAppResetLogin):
    """
    Test how the list of credential changes, when one is added or removed, and after a reset.
    """
    cred1 = b"TESTCRED"
    cred2 = b"ANOTHERCRED"

    secretsApp = secretsAppResetLogin

    assert not secretsApp.list()

    secretsApp.verify_pin_raw(PIN)
    secretsApp.register(cred1, SECRET, DIGITS, kind=Kind.Hotp)
    secretsApp.verify_pin_raw(PIN)
    assert cred1 in secretsApp.list()
    secretsApp.verify_pin_raw(PIN)
    secretsApp.register(cred2, SECRET, DIGITS, kind=Kind.Hotp)
    secretsApp.verify_pin_raw(PIN)
    assert cred2 in secretsApp.list()

    secretsApp.verify_pin_raw(PIN)
    secretsApp.delete(cred2)
    secretsApp.verify_pin_raw(PIN)
    assert cred2 not in secretsApp.list()
    secretsApp.verify_pin_raw(PIN)
    assert cred1 in secretsApp.list()

    secretsApp.reset()
    secretsApp.set_pin_raw(PIN)
    secretsApp.verify_pin_raw(PIN)
    assert not secretsApp.list()


@pytest.mark.parametrize(
    "secret",
    [
        "3132333435363738393031323334353637383930",
        "00" * 19 + "ff",
        "002EF43F51AFA97BA2B46418768123C9E1809A5B",
        "002EF43F51AFA97BA2B46418768123C9E1809A5B" * 2,
    ],
)
@pytest.mark.parametrize(
    "start_counter",
    [
        0,
        0xFF + 1,
        0xFFFF + 1,
        0xFFFFFF + 1,
        0xFFFFFFFF - 10,
    ],
)
def test_calculated_codes_hotp(secretsAppResetLogin, secret, start_counter):
    """
    Test HOTP codes against another OTP library.
    Use different secret and start counter values.
    """
    oath = pytest.importorskip("oath")
    secretb = binascii.a2b_hex(secret)
    secretsApp = secretsAppResetLogin
    secretsApp.register(
        CREDID,
        secretb,
        digits=6,
        kind=Kind.Hotp,
        algo=Algorithm.Sha1,
        initial_counter_value=start_counter,
    )
    lib_at = lambda t: oath.hotp(secret, counter=t, format="dec6").encode()
    for i in range(10):
        i = i + start_counter
        secretsApp.verify_pin_raw(PIN)
        assert secretsApp.calculate(CREDID, i) == lib_at(i)


@pytest.mark.parametrize(
    "secret",
    [
        "3132333435363738393031323334353637383930",
        "00" * 19 + "ff",
        "002EF43F51AFA97BA2B46418768123C9E1809A5B",
        "002EF43F51AFA97BA2B46418768123C9E1809A5B" * 2,
    ],
)
def test_calculated_codes_totp(secretsAppResetLogin, secret):
    """
    Test TOTP codes against another OTP library.
    """
    oath = pytest.importorskip("oath")
    secretb = binascii.a2b_hex(secret)
    secretsApp = secretsAppResetLogin
    secretsApp.register(CREDID, secretb, digits=6, kind=Kind.Totp, algo=Algorithm.Sha1)
    lib_at = lambda t: oath.totp(secret, format="dec6", period=30, t=t * 30).encode()
    for i in range(10):
        secretsApp.verify_pin_raw(PIN)
        assert secretsApp.calculate(CREDID, i) == lib_at(i)


def test_calculated_codes_test_vector(secretsAppResetLogin):
    """
    Check output against RFC4226 test vectors, as provided in
    https://www.rfc-editor.org/rfc/rfc4226#page-32
    """
    secret = "3132333435363738393031323334353637383930"
    secretb = binascii.a2b_hex(secret)

    test_vectors = """
       Count    Hexadecimal    Decimal        HOTP
       0        4c93cf18       1284755224     755224
       1        41397eea       1094287082     287082
       2         82fef30        137359152     359152
       3        66ef7655       1726969429     969429
       4        61c5938a       1640338314     338314
       5        33c083d4        868254676     254676
       6        7256c032       1918287922     287922
       7         4e5b397         82162583     162583
       8        2823443f        673399871     399871
       9        2679dc69        645520489     520489"""
    # select last column only, starting after the header line
    codes = [x.split()[-1].encode() for x in test_vectors.splitlines()[2:]]
    secretsApp = secretsAppResetLogin

    secretsApp.register(CREDID, secretb, digits=6, kind=Kind.Hotp, algo=Algorithm.Sha1)
    for i in range(10):
        secretsApp.verify_pin_raw(PIN)
        assert secretsApp.calculate(CREDID, i) == codes[i]


def test_reverse_hotp_vectors(secretsAppResetLogin):
    """
    Test passing conditions for the HOTP reverse check
    Check against RFC4226 test vectors, as provided in
    https://www.rfc-editor.org/rfc/rfc4226#page-32
    """
    secret = "3132333435363738393031323334353637383930"
    secretb = binascii.a2b_hex(secret)

    test_vectors = """
       Count    Hexadecimal    Decimal        HOTP
       0        4c93cf18       1284755224     755224
       1        41397eea       1094287082     287082
       2         82fef30        137359152     359152
       3        66ef7655       1726969429     969429
       4        61c5938a       1640338314     338314
       5        33c083d4        868254676     254676
       6        7256c032       1918287922     287922
       7         4e5b397         82162583     162583
       8        2823443f        673399871     399871
       9        2679dc69        645520489     520489"""
    # select last column only, starting after the header line
    codes = [x.split()[-1].encode() for x in test_vectors.splitlines()[2:]]

    secretsApp = secretsAppResetLogin
    secretsApp.register(
        CREDID, secretb, digits=6, kind=Kind.HotpReverse, algo=Algorithm.Sha1
    )
    for i in range(10):
        c = int(codes[i])
        secretsApp.verify_pin_raw(PIN)
        assert secretsApp.verify_code(CREDID, c)


def test_reverse_hotp_failure(secretsAppRaw):
    """
    Test failing conditions for the HOTP reverse check
    """
    secret = "3132333435363738393031323334353637383930"
    secretb = binascii.a2b_hex(secret)

    codes = [x for x in range(10)]

    secretsApp = secretsAppRaw
    secretsApp.reset()
    secretsApp.register(
        CREDID, secretb, digits=6, kind=Kind.HotpReverse, algo=Algorithm.Sha1
    )
    # Make sure the obligatory delay has passed in case the previous test has triggered it
    helper_wait_after_failed_hotp_verification_request()
    for i in range(3):
        c = codes[i]
        with pytest.raises(SecretsAppException, match="VerificationFailed"):
            assert not secretsApp.verify_code(CREDID, c)
        helper_wait_after_failed_hotp_verification_request()

    # Test parsing too long code
    with pytest.raises(SecretsAppException, match="VerificationFailed"):
        assert not secretsApp.verify_code(CREDID, 10**5)
    helper_wait_after_failed_hotp_verification_request()

    secretsApp.delete(CREDID)
    secretsApp.register(CREDID, secretb, digits=7, kind=Kind.Hotp, algo=Algorithm.Sha1)
    with pytest.raises(SecretsAppException, match="ConditionsOfUseNotSatisfied"):
        assert not secretsApp.verify_code(CREDID, 10**6)
    helper_wait_after_failed_hotp_verification_request()

    secretsApp.delete(CREDID)
    secretsApp.register(CREDID, secretb, digits=8, kind=Kind.Hotp, algo=Algorithm.Sha1)
    with pytest.raises(SecretsAppException, match="ConditionsOfUseNotSatisfied"):
        assert not secretsApp.verify_code(CREDID, 10**7)
    helper_wait_after_failed_hotp_verification_request()


def helper_wait(seconds: int) -> None:
    l = logging.getLogger()
    l.debug(f"Waiting {seconds} seconds before continuing")
    time.sleep(seconds)


def helper_wait_after_failed_hotp_verification_request():
    if FEATURE_BRUTEFORCE_PROTECTION_ENABLED:
        helper_wait(DELAY_AFTER_FAILED_REQUEST_SECONDS)


@pytest.mark.parametrize(
    "start_value",
    [0, 0xFFFF, 0xFFFFFFFF - HOTP_WINDOW_SIZE - 2],
)
@pytest.mark.parametrize(
    "offset",
    [0, 1, HOTP_WINDOW_SIZE - 1, HOTP_WINDOW_SIZE, HOTP_WINDOW_SIZE + 1],
)
def test_reverse_hotp_window(secretsAppResetLogin, offset, start_value):
    """
    Test reverse HOTP code calculation synchronization.
    Solution contains a means to avoid desynchronization between the host's and device's counters. Device calculates
    up to 9 values ahead of its current counter to find the matching code (in total it calculates HOTP code for 10
    subsequent counter positions). In case:

     - no code would match - the on-device counter will not be changed;
     - code would match, but with some counter's offset (up to 9) - the on-device counter will be set to matched code-generated HOTP counter and incremented by 1;
     - code would match, and the code matches counter without offset - the counter will be incremented by 1;
     - counter overflows while calculating the code within the specified window - error is returned, and in that case a new credential with reset counter should be registered

    Device will stop verifying the HOTP codes in case, when the difference between the host and on-device counters
    will be greater or equal to 10.
    See https://github.com/Nitrokey/nitrokey-hotp-verification#verifying-hotp-code for more information.
    """
    oath = pytest.importorskip("oath")
    secret = "3132333435363738393031323334353637383930"
    secretb = binascii.a2b_hex(secret)
    secretsApp = secretsAppResetLogin
    secretsApp.register(
        CREDID,
        secretb,
        digits=6,
        kind=Kind.HotpReverse,
        algo=Algorithm.Sha1,
        initial_counter_value=start_value,
    )
    lib_at = lambda t: oath.hotp(secret, counter=t, format="dec6").encode()
    code_to_send = lib_at(start_value + offset)
    code_to_send = int(code_to_send)
    if offset > HOTP_WINDOW_SIZE:
        # calls with offset bigger than HOTP_WINDOW_SIZE should fail
        secretsApp.verify_pin_raw(PIN)
        with pytest.raises(SecretsAppException, match="VerificationFailed"):
            secretsApp.verify_code(CREDID, code_to_send)
        helper_wait_after_failed_hotp_verification_request()
    else:
        # check if this code will be accepted on the given offset
        secretsApp.verify_pin_raw(PIN)
        assert secretsApp.verify_code(CREDID, code_to_send)
        # the same code should not be accepted again, unless counted got saturated
        is_counter_saturated = (
            start_value == (0xFFFFFFFF - HOTP_WINDOW_SIZE)
            and offset == HOTP_WINDOW_SIZE
        )
        if not is_counter_saturated:
            with pytest.raises(
                SecretsAppException,
                match="UnspecifiedPersistentExecutionError|VerificationFailed",
            ):
                secretsAppResetLogin.verify_pin_raw(PIN)
                # send the same code once again - should be rejected
                secretsApp.verify_code(CREDID, code_to_send)
            helper_wait_after_failed_hotp_verification_request()
            # test the very next value - should be accepted
            code_to_send = lib_at(start_value + offset + 1)
            code_to_send = int(code_to_send)
            secretsAppResetLogin.verify_pin_raw(PIN)
            assert secretsApp.verify_code(CREDID, code_to_send)
        else:
            # counter got saturated, error code will be returned
            for _ in range(3):
                with pytest.raises(
                    SecretsAppException, match="UnspecifiedPersistentExecutionError"
                ):
                    secretsAppResetLogin.verify_pin_raw(PIN)
                    secretsApp.verify_code(CREDID, code_to_send)
                helper_wait_after_failed_hotp_verification_request()


@pytest.mark.parametrize(
    "digits",
    [6, 8],
)
@pytest.mark.parametrize(
    "algorithm",
    [
        (Algorithm.Sha1, hashlib.sha1),
        (Algorithm.Sha256, hashlib.sha256),
        # (Algorithm.Sha512, hashlib.sha512),  # unsupported by the OTP App in the firmware
    ],
)
@pytest.mark.parametrize(
    "secret",
    [
        "3132333435363738393031323334353637383930",
        "002EF43F51AFA97BA2B46418768123C9E1809A5B" * 2,
    ],
)
def test_calculated_codes_totp_hash_digits(
    secretsAppResetLogin, secret, algorithm, digits
):
    """
    Test TOTP codes against another OTP library, with different hash algorithms and digits count.
    Test vector secret, and a random 40 bytes value.
    """
    algo_app, algo_oath = algorithm
    oath = pytest.importorskip("oath")
    secretb = binascii.a2b_hex(secret)
    secretsApp = secretsAppResetLogin
    secretsApp.register(CREDID, secretb, digits=digits, kind=Kind.Totp, algo=algo_app)
    lib_at = lambda t: oath.totp(
        secret, format="dec" + str(digits), period=30, t=t * 30, hash=algo_oath
    ).encode()
    for i in range(10):
        secretsApp.verify_pin_raw(PIN)
        assert secretsApp.calculate(CREDID, i) == lib_at(i)


@pytest.mark.slow
@pytest.mark.parametrize(
    "long_labels",
    ["short_labels", "long_labels"],
)
@pytest.mark.parametrize(
    "kind",
    [Kind.Totp, Kind.Hotp],
)
@pytest.mark.parametrize(
    "count",
    [
        30,
        pytest.param(1000, marks=pytest.mark.slow),
    ],
)
def test_load(secretsAppResetLogin, kind: Kind, long_labels: str, count):
    """
    Load tests to see how much OTP credentials we can store,
    and if using of them is not broken with the full FS.
    """
    secret = "3132333435363738393031323334353637383930"
    oath = pytest.importorskip("oath")
    secretb = binascii.a2b_hex(secret)

    secretsApp = secretsAppResetLogin
    credentials_registered: int = 0
    names_registered: List[bytes] = []

    name_gen: Callable[[int], str] = lambda x: f"LOAD{x:02}"
    if long_labels == "long_labels":
        name_gen = lambda x: (f"LOAD{x:02}" * 100)[:CREDENTIAL_LABEL_MAX_SIZE]

    i = 0
    while True:
        name = name_gen(i)
        try:
            secretsApp.verify_pin_raw(PIN)
            secretsApp.register(
                name.encode(), secretb, digits=6, kind=kind, initial_counter_value=i
            )
            names_registered.append(name.encode())
            if i > count:
                i = i + 1
                raise Exception("Reached expected credentials count, finishing early")
        except Exception as e:
            print(f"{e}")
            print(f"Registered {i} credentials")
            size = len(secret) + len(name)
            print(f"Single Credential size: {size} B")
            print(f"Total size: {size * i} B")
            credentials_registered = i
            break
        i += 1

    assert (
        credentials_registered > 30
    ), "Expecting being able to register at least 30 OTP credentials"

    secretsApp.verify_pin_raw(PIN)
    l = secretsApp.list()
    assert sorted(l) == sorted(names_registered)
    assert len(l) == credentials_registered

    # Make some space for the counter updates - delete the last 3 credentials
    CRED_TO_REMOVE = 3
    for name_c in names_registered[-CRED_TO_REMOVE:]:
        secretsApp.verify_pin_raw(PIN)
        secretsApp.delete(name_c)
    credentials_registered -= CRED_TO_REMOVE

    secretsApp.verify_pin_raw(PIN)
    l = secretsApp.list()
    assert len(l) == credentials_registered

    lib_at = lambda t: oath.totp(secret, format="dec6", period=30, t=t * 30).encode()
    if kind == Kind.Hotp:
        lib_at = lambda t: oath.hotp(secret, format="dec6", counter=t).encode()

    for i in range(credentials_registered):
        # At this point device should respond to our calls, despite being full, fail otherwise
        # Iterate over credentials and check code at given challenge
        nameb = name_gen(i).encode()
        secretsApp.verify_pin_raw(PIN)
        assert secretsApp.calculate(nameb, i) == lib_at(i)

    secretsApp.verify_pin_raw(PIN)
    l = secretsApp.list()
    assert len(l) == credentials_registered


def test_remove_all_credentials_by_hand(secretsAppRaw):
    """
    Remove all hold credentials by hand and test for being empty.
    """
    secretsApp = secretsAppRaw
    secretsApp.reset()
    secretsApp.set_pin_raw(PIN)
    cred_no_pbek = [f"CredNoPBEK{i}".encode() for i in range(1, 5)]
    cred_pbek = [f"CredPBEK{i}".encode() for i in range(1, 5)]
    all_registered = sorted([*cred_no_pbek, *cred_pbek])

    for c in cred_no_pbek:
        secretsApp.register(
            c, SECRET, DIGITS, pin_based_encryption=False, kind=Kind.Hotp
        )

    for c in cred_pbek:
        secretsApp.verify_pin_raw(PIN)
        secretsApp.register(
            c, SECRET, DIGITS, pin_based_encryption=True, kind=Kind.Hotp
        )

    secretsApp.verify_pin_raw(PIN)
    credential_list = secretsApp.list()
    assert len(credential_list) > 0, "Empty credentials list"
    assert sorted(secretsApp.list()) == cred_no_pbek
    secretsApp.verify_pin_raw(PIN)
    assert sorted(secretsApp.list()) == all_registered

    # Delete all listed credentials by hand
    for credential_name in credential_list:
        try:
            secretsApp.delete(credential_name)
        except SecretsAppException as e:
            if "NotFound" in str(e):
                # Try again with provided PIN
                secretsApp.verify_pin_raw(PIN)
                secretsApp.delete(credential_name)

    secretsApp.verify_pin_raw(PIN)
    credential_list = secretsApp.list()
    assert len(credential_list) == 0


def test_send_rubbish(secretsAppRaw):
    """Check if the application crashes, when sending unexpected data for the given command"""
    secretsApp = secretsAppRaw
    secretsApp.reset()
    secretsApp.register(CREDID, SECRET, DIGITS, kind=Kind.Hotp)

    # Just randomly selected 20 bytes of non-TLV data
    invalid_data = bytes([0x11] * 20)
    for _ in range(3):
        with pytest.raises(fido2.ctap.CtapError):
            secretsApp._send_receive_inner(invalid_data)
    secretsApp.list()

    # Skip some commands from test
    for ins in set(Instruction).difference(
        {
            # These commands do not parse
            Instruction.Reset,
            Instruction.List,
            Instruction.SendRemaining,
            # Following are deactivated
            *CHALLENGE_RESPONSE_COMMANDS,
            *CALCULATE_ALL_COMMANDS,
        }
    ):
        with pytest.raises(
            Exception,
            match="CTAP error|IncorrectDataParameter|InstructionNotSupportedOrInvalid",
        ):
            structure = [
                RawBytes([0x02, 0x02]),
            ]
            secretsApp._send_receive(ins, structure)
    secretsApp.list()


def test_too_long_message(secretsAppResetLogin):
    """
    Check device's response for the too long message
    """
    secretsApp = secretsAppResetLogin
    secretsApp.register(CREDID, SECRET, DIGITS, kind=Kind.Hotp)
    secretsApp.verify_pin_raw(PIN)
    secretsApp.list()

    too_long_name = b"a" * 253
    with pytest.raises(SecretsAppException, match="IncorrectDataParameter"):
        structure = [
            tlv8.Entry(Tag.CredentialId.value, too_long_name),
        ]
        secretsApp._send_receive(Instruction.Put, structure)
    secretsApp.verify_pin_raw(PIN)
    secretsApp.list()


def test_too_long_message2(secretsAppRaw):
    """
    Test how long the secret could be
    """
    secretsApp = secretsAppRaw
    secretsApp.reset()
    secretsApp.register(CREDID, SECRET, DIGITS, kind=Kind.Hotp)
    secretsApp.list()

    # Check maximum label length
    too_long_name = b"a" * 256
    additional_space = 100
    max_label_length = len(SECRET) + additional_space
    secretsApp.register(
        too_long_name[:-max_label_length], SECRET, DIGITS, kind=Kind.Hotp
    )

    # Find out experimentally the maximum accepted secret length - 126 bytes
    # Use minimal label length
    codes = []
    i = 0
    for i in range(1, 255, 5):
        secretsApp.logfn(f"Testing secret length {i} bytes")
        try:
            with suppress(SecretsAppException):
                secretsAppRaw.delete(CREDID)
            secretsApp.register(CREDID, too_long_name[:i], DIGITS, kind=Kind.Hotp)
            codes.append(secretsApp.calculate(CREDID))
        except Exception:
            break
    assert i >= 40, "Maximum secret length should be at least 320 bits"
    assert len(list(set(codes))) == len(
        codes
    ), "All returned OTP codes should be unique"


def test_status(secretsApp):
    """
    Simple test for getting device's status
    """
    print(secretsApp.select())


@pytest.mark.skipif(
    not FEATURE_CHALLENGE_RESPONSE_ENABLED,
    reason="Challenge-Response feature should be activated",
)
def test_set_code(secretsApp):
    """
    Simple test for setting the proper code on the device.
    """
    SECRET = b"1" * 20
    CHALLENGE = b"1234"

    secretsApp.reset()
    state = secretsApp.select()
    print(state)
    assert state.algorithm is None
    assert state.challenge is None

    response = hmac.HMAC(key=SECRET, msg=CHALLENGE, digestmod="sha1").digest()
    secretsApp.set_code_raw(SECRET, CHALLENGE, response)

    state = secretsApp.select()
    print(state)
    assert state.challenge is not None
    assert state.algorithm is not None


@pytest.mark.parametrize(
    "remove_password_with",
    [
        Instruction.Reset,
        Instruction.SetCode,
    ],
)
@pytest.mark.skipif(
    not FEATURE_CHALLENGE_RESPONSE_ENABLED,
    reason="Challenge-Response feature should be activated",
)
def test_set_code_and_validate(secretsAppRaw, remove_password_with: Instruction):
    """
    Test device's behavior when the validation code is set.
    Non-authorized calls should be rejected, except for the selected.
    Authorization should be valid only until the next call.

    Authorization is needed for all the listed commands, except for RESET and VALIDATE:
         Required               Not required
         PUT 0x01               RESET 0x04 N
         DELETE 0x02            VALIDATE 0xa3 N
         SET CODE 0x03
         LIST 0xa1
         CALCULATE 0xa2
         CALCULATE ALL 0xa4
         SEND REMAINING 0xa5
    Details:
    - https://developers.yubico.com/OATH/YKOATH_Protocol.html
    """

    # The secret in production should be:
    #   SECRET = PBKDF2(USER_PASSPHRASE || DEVICEID, 1000)[:16]
    SECRET = b"1" * 20
    CHALLENGE = b"12345678"  # in production should be random 8 bytes

    # Device should be in the non-protected mode, and list command is allowed
    secretsApp = secretsAppRaw
    secretsApp.reset()
    secretsApp.list()

    # Set the code, and require validation before regular calls from now on
    response = hmac.HMAC(key=SECRET, msg=CHALLENGE, digestmod="sha1").digest()
    secretsApp.set_code_raw(SECRET, CHALLENGE, response)

    # Make sure all the expected commands are failing, as in specification
    with pytest.raises(SecretsAppException, match="ConditionsOfUseNotSatisfied"):
        secretsApp.list()

    for ins in set(Instruction) - {Instruction.Reset, Instruction.Validate}:
        with pytest.raises(
            SecretsAppException,
            match="IncorrectDataParameter|InstructionNotSupportedOrInvalid|NotFound|ConditionsOfUseNotSatisfied",
        ):
            structure = [RawBytes([0x02] * 10)]
            secretsApp._send_receive(ins, structure)

    # Each guarded command has to prepended by the validation call
    # Run "list" command, with validation first
    state = secretsApp.select()
    response_validate = hmac.HMAC(
        key=SECRET, msg=state.challenge, digestmod="sha1"
    ).digest()
    secretsApp.validate_raw(challenge=state.challenge, response=response_validate)
    secretsApp.list()

    # Make sure another command call is not allowed
    with pytest.raises(SecretsAppException, match="ConditionsOfUseNotSatisfied"):
        secretsApp.list()

    # Test running "list" command again
    state = secretsApp.select()
    response_validate = hmac.HMAC(
        key=SECRET, msg=state.challenge, digestmod="sha1"
    ).digest()
    secretsApp.validate_raw(challenge=state.challenge, response=response_validate)
    secretsApp.list()

    if remove_password_with == Instruction.Reset:
        # Reset should be allowed
        secretsApp.reset()
    elif remove_password_with == Instruction.SetCode:
        # Clearing passphrase should be allowed after authentication
        with pytest.raises(SecretsAppException, match="ConditionsOfUseNotSatisfied"):
            secretsApp.clear_code()

        state = secretsApp.select()
        response_validate = hmac.HMAC(
            key=SECRET, msg=state.challenge, digestmod="sha1"
        ).digest()
        secretsApp.validate_raw(challenge=state.challenge, response=response_validate)
        secretsApp.clear_code()
    else:
        raise ValueError()

    state = secretsApp.select()
    assert state.challenge is None


@pytest.mark.skip(reason="This test takes long time")
def test_revhotp_bruteforce(secretsAppNoLog):
    """
    This test implements practical brute-forcing of the codes values.
    In case multiple devices use the same secret, stealing and brute-forcing answers on one
    could help with the other.
    """
    secretsApp = secretsAppNoLog
    secretsApp.reset()
    secretsApp.register(
        CREDID, SECRET, digits=6, kind=Kind.HotpReverse, algo=Algorithm.Sha1
    )
    start_time = time.time()
    code_start = 1_000_000

    from tqdm import tqdm, trange

    for current_code in trange(code_start, 0, -1):
        tqdm.write(f"Trying code {current_code}")
        try:
            secretsApp.verify_code(CREDID, current_code)
            stop_time = time.time()
            tqdm.write(
                f"Found code {current_code} after {stop_time - start_time} seconds"
            )
            break
        except KeyboardInterrupt:
            break
        except fido2.ctap.CtapError:
            pass
        except Exception:
            break


@pytest.mark.skipif(
    FEATURE_BRUTEFORCE_PROTECTION_ENABLED == False,
    reason="Brute-force protection feature should be activated",
)
def test_revhotp_delay_on_failure(secretsAppRaw):
    """
    Check if the right delay is set, when the invalid code is given for the reverse HOTP operation.
    On failure the response time should take at least 1 second to prevent easy brute force.
    """

    secretsApp = secretsAppRaw
    intentionally_wrong_code = 123123

    secretsApp.reset()
    secretsApp.register(
        CREDID, SECRET, digits=6, kind=Kind.HotpReverse, algo=Algorithm.Sha1
    )
    start_time = time.time()
    stop_time = start_time + DELAY_AFTER_FAILED_REQUEST_SECONDS
    with pytest.raises(SecretsAppException, match="VerificationFailed"):
        secretsApp.verify_code(CREDID, intentionally_wrong_code)
    while time.time() < stop_time:
        with pytest.raises(SecretsAppException, match="SecurityStatusNotSatisfied"):
            secretsApp.verify_code(CREDID, intentionally_wrong_code)
        helper_wait(DELAY_AFTER_FAILED_REQUEST_SECONDS / 10)
    with pytest.raises(SecretsAppException, match="VerificationFailed"):
        helper_wait(DELAY_AFTER_FAILED_REQUEST_SECONDS / 10)
        secretsApp.verify_code(CREDID, intentionally_wrong_code)


def test_set_pin(secretsApp):
    """
    Simple test for setting the PIN on the device.
    """
    secretsApp.reset()
    state = secretsApp.select()
    print(state)
    assert state.algorithm is None
    assert state.challenge is None
    assert state.pin_attempt_counter is None

    secretsApp.set_pin_raw(PIN)

    state = secretsApp.select()
    print(state)
    assert state.challenge is None
    assert state.algorithm is None
    assert state.pin_attempt_counter == PIN_ATTEMPT_COUNTER_DEFAULT

    # Should fail when setting the second time
    with pytest.raises(SecretsAppException, match="SecurityStatusNotSatisfied"):
        secretsApp.set_pin_raw(PIN)


def test_change_pin(secretsApp):
    """
    Simple test for setting the proper code on the device.
    """
    secretsApp.reset()

    state = secretsApp.select()
    assert state.pin_attempt_counter is None

    secretsApp.set_pin_raw(PIN)
    assert secretsApp.select().pin_attempt_counter == PIN_ATTEMPT_COUNTER_DEFAULT

    secretsApp.change_pin_raw(PIN, PIN2)
    assert secretsApp.select().pin_attempt_counter == PIN_ATTEMPT_COUNTER_DEFAULT

    secretsApp.change_pin_raw(PIN2, PIN)
    assert secretsApp.select().pin_attempt_counter == PIN_ATTEMPT_COUNTER_DEFAULT

    # Should fail when setting the second time with the PIN2
    with pytest.raises(SecretsAppException, match="VerificationFailed"):
        secretsApp.change_pin_raw(PIN2, PIN)
    # after providing the wrong PIN, the attempt counter should decrement itself by 1
    assert secretsApp.select().pin_attempt_counter == PIN_ATTEMPT_COUNTER_DEFAULT - 1


def test_change_pin_data_dont_change(secretsAppResetLogin):
    """
    Test for changing the proper PIN on the device.
    Check if data remain the same, both PIN and Hardware key encrypted
    """

    def helper_test_calculated_codes_totp(secretsApp, secret: str, PIN: str):
        """Test TOTP codes against another OTP library."""
        oath = pytest.importorskip("oath")
        lib_at = lambda t: oath.totp(
            secret, format="dec6", period=30, t=t * 30
        ).encode()
        for i in range(10):
            # Use non-modified verify_pin_raw_always call to always verify PIN, regardless of the fixture type
            secretsApp.verify_pin_raw_always(PIN)
            assert secretsApp.calculate(CREDID, i) == lib_at(i)

    # Initial setup for the TOTP slot and test
    secret = "00" * 20
    secretsApp = secretsAppResetLogin
    secretb = binascii.a2b_hex(secret)
    secretsApp.register(CREDID, secretb, digits=6, kind=Kind.Totp, algo=Algorithm.Sha1)
    helper_test_calculated_codes_totp(secretsApp, secret, PIN)

    # Change PIN and check whether the data remain the same
    secretsApp.change_pin_raw(PIN, PIN2)
    assert secretsApp.select().pin_attempt_counter == PIN_ATTEMPT_COUNTER_DEFAULT
    helper_test_calculated_codes_totp(secretsApp, secret, PIN2)

    # And again
    secretsApp.change_pin_raw(PIN2, PIN)
    assert secretsApp.select().pin_attempt_counter == PIN_ATTEMPT_COUNTER_DEFAULT
    helper_test_calculated_codes_totp(secretsApp, secret, PIN)

    # Should fail when setting the second time with the PIN2
    with pytest.raises(SecretsAppException, match="VerificationFailed"):
        secretsApp.change_pin_raw(PIN2, PIN)

    # After providing the wrong PIN, the attempt counter should decrement itself by 1
    assert secretsApp.select().pin_attempt_counter == PIN_ATTEMPT_COUNTER_DEFAULT - 1

    # After failed PIN change the data should remain the same
    helper_test_calculated_codes_totp(secretsApp, secret, PIN)

    # After providing the correct PIN, the attempt counter should reset itself
    assert secretsApp.select().pin_attempt_counter == PIN_ATTEMPT_COUNTER_DEFAULT


def test_verify_pin(secretsApp):
    """
    Simple test for PIN verification
    """
    secretsApp.reset()
    secretsApp.set_pin_raw(PIN)
    secretsApp.verify_pin_raw(PIN)
    secretsApp.list()
    assert secretsApp.select().pin_attempt_counter == PIN_ATTEMPT_COUNTER_DEFAULT

    # Make sure all the expected commands are working, as in specification
    secretsApp.list()

    # With PIN verified this should work as well
    secretsApp.verify_pin_raw(PIN)
    secretsApp.list()


def test_use_up_pin_counter(secretsAppRaw):
    secretsApp = secretsAppRaw
    secretsApp.reset()
    secretsApp.set_pin_raw(PIN)
    assert not secretsApp.list()
    assert secretsApp.select().pin_attempt_counter == PIN_ATTEMPT_COUNTER_DEFAULT

    cred_no_pbek = [f"CredNoPBEK{i}".encode() for i in range(1, 5)]
    cred_pbek = [f"CredPBEK{i}".encode() for i in range(1, 5)]

    for i, c in enumerate(cred_no_pbek):
        secretsApp.register(
            c,
            SECRET,
            DIGITS,
            initial_counter_value=i,
            pin_based_encryption=False,
            kind=Kind.Hotp,
        )
    assert sorted(secretsApp.list()) == cred_no_pbek

    for c in cred_pbek:
        secretsApp.verify_pin_raw(PIN)
        secretsApp.register(
            c, SECRET, DIGITS, pin_based_encryption=True, kind=Kind.Totp
        )
    assert sorted(secretsApp.list()) == cred_no_pbek

    # Use all PIN counter attempts
    while secretsApp.select().pin_attempt_counter > 0:
        with pytest.raises(SecretsAppException, match="VerificationFailed"):
            secretsApp.verify_pin_raw(PIN2)

    # With the PIN attempt counter used up, verifying with the correct PIN should not recover the device
    assert secretsApp.select().pin_attempt_counter == 0
    with pytest.raises(SecretsAppException, match="VerificationFailed"):
        secretsApp.verify_pin_raw(PIN)
    assert secretsApp.select().pin_attempt_counter == 0

    # Standard commands should not require authentication
    # Check if non-encrypted are still there and usable
    assert sorted(secretsApp.list()) == cred_no_pbek
    for c in cred_no_pbek:
        secretsApp.calculate(c)

    assert secretsApp.select().pin_attempt_counter == 0
    with pytest.raises(SecretsAppException, match="VerificationFailed"):
        secretsApp.verify_pin_raw(PIN)
    for c in cred_pbek:
        with pytest.raises(SecretsAppException, match="NotFound"):
            secretsApp.calculate(c)

    secretsApp.reset()
    secretsApp.set_pin_raw(PIN)
    assert not secretsApp.list()
    assert secretsApp.select().pin_attempt_counter == PIN_ATTEMPT_COUNTER_DEFAULT


def test_list_pin_no_pin(secretsAppRaw):
    """
    Check list command behavior in the presence of PIN and without it
    1. Credential list should be empty after reset
    2. Credential list should show only non-PIN-encrypted credentials, when unauthorized
    3. Credential list should show both non- and PIN-encrypted credentials, when authorized
    @param secretsAppRaw: Direct access to the Secret App API, without any additional setup
    """
    secretsApp = secretsAppRaw

    secretsApp.reset()
    secretsApp.set_pin_raw(PIN)
    secretsApp.verify_pin_raw(PIN)
    # 1. Credential list should be empty after reset
    assert not secretsApp.list()

    cred_no_pbek = [f"CredNoPBEK{i}".encode() for i in range(1, 5)]
    cred_pbek = [f"CredPBEK{i}".encode() for i in range(1, 5)]

    for c in cred_no_pbek:
        secretsApp.register(
            c, SECRET, DIGITS, pin_based_encryption=False, kind=Kind.Hotp
        )

    assert sorted(secretsApp.list()) == cred_no_pbek

    for c in cred_pbek:
        secretsApp.verify_pin_raw(PIN)
        secretsApp.register(
            c, SECRET, DIGITS, pin_based_encryption=True, kind=Kind.Hotp
        )

    # 2. Credential list should show only non-PIN-encrypted credentials, when unauthorized
    assert sorted(secretsApp.list()) == cred_no_pbek

    # 3. Credential list should show both non- and PIN-encrypted credentials, when authorized
    secretsApp.verify_pin_raw(PIN)
    assert sorted(secretsApp.list()) == sorted([*cred_no_pbek, *cred_pbek])

    # 2. Credential list should show only non-PIN-encrypted credentials, when unauthorized
    assert sorted(secretsApp.list()) == cred_no_pbek

    # 1. Credential list should be empty after reset
    secretsApp.reset()
    secretsApp.set_pin_raw(PIN)
    assert not secretsApp.list()
    secretsApp.verify_pin_raw(PIN)
    assert not secretsApp.list()


def test_check_commands_access(secretsAppRaw):
    """
    Access to all commands should be always available. Sending garbage, and expecting parsing error.
    If this was access guarded, the parsing would not happen.
    """

    # Not tested with challenge-response enabled
    assert not FEATURE_CHALLENGE_RESPONSE_ENABLED

    for ins in set(Instruction) - {
        Instruction.CalculateAll,
        *CHALLENGE_RESPONSE_COMMANDS,
    }:
        structure = [RawBytes([0x02] * 10)]
        if ins not in [Instruction.Reset, Instruction.List, Instruction.SendRemaining]:
            # Check if we get parsing error from these
            with pytest.raises(SecretsAppException, match="IncorrectDataParameter"):
                secretsAppRaw._send_receive(ins, structure)
        elif ins == Instruction.SendRemaining:
            # This one should not be allowed without data awaiting
            with pytest.raises(
                SecretsAppException, match="ConditionsOfUseNotSatisfied"
            ):
                secretsAppRaw._send_receive(ins, structure)
        else:
            # These commands do not expect arguments, so there is no parsing error
            secretsAppRaw._send_receive(ins, structure)


def test_register_pin_encrypted_without_auth(secretsAppRaw):
    """
    PIN-encrypted Credential registration should not be possible without previous verification
    """
    secretsAppRaw.reset()
    with pytest.raises(SecretsAppException, match="SecurityStatusNotSatisfied"):
        secretsAppRaw.register(
            CREDID, SECRET, DIGITS, pin_based_encryption=True, kind=Kind.Hotp
        )
    assert not secretsAppRaw.list()

    secretsAppRaw.set_pin_raw(PIN)
    secretsAppRaw.verify_pin_raw(PIN)
    secretsAppRaw.register(
        CREDID, SECRET, DIGITS, pin_based_encryption=True, kind=Kind.Hotp
    )
    secretsAppRaw.verify_pin_raw(PIN)
    assert secretsAppRaw.list()


def test_pin_operations_do_not_authenticate(secretsAppRaw):
    """
    Make sure that PIN operations other than Verify do not authenticate
    """
    secretsAppRaw.reset()
    with pytest.raises(SecretsAppException, match="SecurityStatusNotSatisfied"):
        secretsAppRaw.register(
            CREDID, SECRET, DIGITS, pin_based_encryption=True, kind=Kind.Hotp
        )
    assert not secretsAppRaw.list()

    secretsAppRaw.set_pin_raw(PIN)
    with pytest.raises(SecretsAppException, match="SecurityStatusNotSatisfied"):
        secretsAppRaw.register(
            CREDID, SECRET, DIGITS, pin_based_encryption=True, kind=Kind.Hotp
        )
    assert not secretsAppRaw.list()

    secretsAppRaw.change_pin_raw(PIN, PIN2)
    with pytest.raises(SecretsAppException, match="SecurityStatusNotSatisfied"):
        secretsAppRaw.register(
            CREDID, SECRET, DIGITS, pin_based_encryption=True, kind=Kind.Hotp
        )
    assert not secretsAppRaw.list()
    secretsAppRaw.verify_pin_raw(PIN2)
    assert not secretsAppRaw.list()

    secretsAppRaw.verify_pin_raw(PIN2)
    secretsAppRaw.register(
        CREDID, SECRET, DIGITS, pin_based_encryption=True, kind=Kind.Hotp
    )
    secretsAppRaw.verify_pin_raw(PIN2)
    assert secretsAppRaw.list()


def test_credential_encryption_does_not_change(secretsAppRaw):
    """
    Check if counter-updated HOTP Credential does not change its encryption type after use.
    """

    secretsAppRaw.reset()
    assert not secretsAppRaw.list()
    secretsAppRaw.set_pin_raw(PIN)

    # Register PIN-protected credential
    secretsAppRaw.verify_pin_raw(PIN)
    secretsAppRaw.register(
        CREDID, SECRET, DIGITS, pin_based_encryption=True, kind=Kind.Hotp
    )
    assert not secretsAppRaw.list()
    secretsAppRaw.verify_pin_raw(PIN)
    assert secretsAppRaw.list()

    # Credential should not be available for use without presenting the PIN
    with pytest.raises(SecretsAppException, match="NotFound"):
        secretsAppRaw.calculate(CREDID, CHALLENGE)
    secretsAppRaw.verify_pin_raw(PIN)
    secretsAppRaw.calculate(CREDID, CHALLENGE)

    # After use the HOTP counter was updated, and the Credential saved back. The credential should still be PIN-protected.
    with pytest.raises(SecretsAppException, match="NotFound"):
        secretsAppRaw.calculate(CREDID, CHALLENGE)
    assert not secretsAppRaw.list()


def helper_send_receive_ins(
    app: SecretsApp,
    ins,
    structure=None,
    p1=None,
    le=None,
    data_raw: bytes = b"",
    expected_SW: Optional[str] = "9000",
) -> Tuple[bytes, bytes]:
    """
    Helper for direct communication with the device
    Send data directly to the CTAP bridge.
    @param app: App
    @param ins: CCID Ins; if sent as int, it is used verbatim, otherwise will be encoded as default
    @param structure: a list of TLV entries to encode
    @param p1: CCID P1 parameter to be called; if not provided, a default for given Ins will be used
    @param le: CCID Le
    @param data_raw: Raw CCID data field to send; used, when structure is None
    @param expected_SW: expected SW code, which is checked before returning
    """

    def _trunc(s: str, l: int = 100) -> str:
        return f"{s[:l]}.." if len(s) > l else s

    from pynitrokey.start.gnuk_token import iso7816_compose

    p2 = 0
    p1 = 0 if p1 is None else p1
    if isinstance(ins, int):
        ins_b = ins
    else:
        ins_b, p1, p2 = app._encode_command(ins)
    data_to_send = app._custom_encode(structure) if structure is not None else data_raw
    data = iso7816_compose(ins_b, p1, p2, data_to_send, le=le)
    app.logfn(f">> {_trunc(data.hex())}")
    res = app.dev.otp(data=data)
    app.logfn(f"<< {_trunc(res.hex())}")
    status_bytes, result = res[:2], res[2:]

    if expected_SW:
        assert status_bytes.hex() == expected_SW
    return status_bytes, result


def test_send_remaining(secretsApp):
    secrets_app = secretsApp
    secrets_app.reset()
    secrets_app.set_pin_raw(PIN)

    with pytest.raises(SecretsAppException, match="ConditionsOfUseNotSatisfied"):
        secrets_app._send_receive(Instruction.SendRemaining)

    # Call list which will have populated "remaining" buffer
    # Call other command
    # The send_remaining call should be denied
    # We need a lot of data to push here, so the ctaphid buffer will be filled (3kB)
    credential_name_length = 127
    count = 3072 // credential_name_length + 1
    secrets_app.logfn(f"Registering {count} credentials")
    credentials = [
        f"Credential{i}".zfill(credential_name_length).encode() for i in range(1, count)
    ]
    for c in credentials:
        secrets_app.verify_pin_raw(PIN)
        secrets_app.register(c, SECRET, DIGITS, kind=Kind.Hotp)
    secrets_app.verify_pin_raw(PIN)
    assert sorted(secrets_app.list()) == sorted(credentials)

    # Run PIN verification so all Credentials on List command will be visible
    secrets_app.verify_pin_raw(PIN)
    status_bytes, result = helper_send_receive_ins(
        secrets_app, Instruction.List, expected_SW=None
    )
    # Make sure there are remaining data to receive
    MORE_DATA_STATUS_BYTE = 0x61
    assert status_bytes[0] == MORE_DATA_STATUS_BYTE
    # Call a different command now, like Delete, which should not add any new data to the buffer
    status_bytes, result = helper_send_receive_ins(
        secrets_app,
        Instruction.Delete,
        structure=[
            tlv8.Entry(Tag.CredentialId.value, credentials[-1]),
        ],
        expected_SW=None,
    )
    assert status_bytes.hex() == "9000"

    # Now that the buffer should be emptied, the SendRemaining call should be rejected
    with pytest.raises(SecretsAppException, match="ConditionsOfUseNotSatisfied"):
        secrets_app._send_receive(Instruction.SendRemaining)


@pytest.mark.parametrize(
    "length",
    [
        # Average daily use case
        30,
        # Maximum field length -> 128B
        128 - 1,
    ],
)
def test_password_safe(secretsAppResetLogin: SecretsApp, length: int) -> None:
    """
    Create a full credential, with both OTP and PWS fields populated. Test working both, with and without PIN-based encryption.
    """
    secretsApp = secretsAppResetLogin
    oath = pytest.importorskip("oath")
    login = b"login".center(length, b"=")
    password = b"password".center(length, b"=")
    metadata = b"metadata".center(length, b"=")
    name = CREDID.center(length, "=").encode()
    secretb = binascii.a2b_hex(SECRET)
    secretsApp.register(
        name,
        secretb,
        digits=6,
        kind=Kind.Totp,
        algo=Algorithm.Sha1,
        login=login,
        password=password,
        metadata=metadata,
    )
    lib_at = lambda t: oath.totp(
        SECRET.decode(), format="dec6", period=30, t=t * 30
    ).encode()
    for i in range(10):
        secretsApp.verify_pin_raw(PIN)
        assert secretsApp.calculate(name, i) == lib_at(i)

    secretsApp.verify_pin_raw(PIN)
    p: PasswordSafeEntry = secretsApp.get_credential(name)
    assert p.name == name
    assert p.login == login
    assert p.password == password
    assert p.metadata == metadata


def test_password_safe_empty_credential(secretsAppResetLogin):
    """
    It should be possible to create an empty credential, with just the name presented
    """
    secretsAppResetLogin.verify_pin_raw(PIN)
    secretsAppResetLogin.register(CREDID)

    with pytest.raises(SecretsAppException, match="ConditionsOfUseNotSatisfied"):
        secretsAppResetLogin.verify_pin_raw(PIN)
        secretsAppResetLogin.calculate(CREDID, 0)

    secretsAppResetLogin.verify_pin_raw(PIN)
    p = secretsAppResetLogin.get_credential(CREDID)
    assert p.name == CREDID.encode()
    assert p.login is None
    assert p.password is None
    assert p.metadata is None

    secretsAppResetLogin.verify_pin_raw(PIN)
    assert CREDID.encode() in secretsAppResetLogin.list()


def test_password_safe_just_pws_entry(secretsAppResetLogin):
    """
    It should be possible to create a PWS-only credential
    """
    length = 20
    login = b"login".center(length, b"=")
    password = b"password".center(length, b"=")
    metadata = b"metadata".center(length, b"=")

    secretsAppResetLogin.verify_pin_raw(PIN)
    secretsAppResetLogin.register(
        CREDID, login=login, password=password, metadata=metadata
    )

    # Since OTP details were not specified, calling Calculate on it should fail
    secretsAppResetLogin.verify_pin_raw(PIN)
    with pytest.raises(SecretsAppException, match="ConditionsOfUseNotSatisfied"):
        secretsAppResetLogin.calculate(CREDID, 0)

    # Reverse HOTP is rejecting that too
    secretsAppResetLogin.verify_pin_raw(PIN)
    with pytest.raises(SecretsAppException, match="ConditionsOfUseNotSatisfied"):
        secretsAppResetLogin.verify_code(CREDID, 0)

    # Let's check what is the content of the credential
    secretsAppResetLogin.verify_pin_raw(PIN)
    p = secretsAppResetLogin.get_credential(CREDID)
    assert p.name == CREDID.encode()
    assert p.login == login
    assert p.password == password
    assert p.metadata == metadata

    secretsAppResetLogin.verify_pin_raw(PIN)
    assert CREDID.encode() in secretsAppResetLogin.list()


def test_select_applet(secretsAppRaw):
    """
    Low-level test for testing the response of the select command
    """
    data = "00 a4 04 00   07 a0 00 00   05 27 21 01"
    data = data.replace(" ", "")
    data = binascii.a2b_hex(data)
    res = secretsAppRaw.dev.otp(data=data)
    assert res.hex() != "6a82"
    assert res.hex().startswith("9000")  # 90007903040a00710869f72b4b3712f627
    print(res.hex())


def helper_get_padded(challenge: bytes, l: int = 64) -> bytes:
    """
    Get PKCS#7 padded buffer
    """
    from cryptography.hazmat.primitives import padding

    # The value passed here is in bits
    padder = padding.PKCS7(l * 8).padder()
    challenge_padded = padder.update(challenge)
    challenge_padded += padder.finalize()
    return challenge_padded


def test_hmac_low_level(secretsAppRaw):
    """
    Test HMAC Challenge setup and use, for KeepassXC support.
    Low-level test.
    Support for this feature is not planned to be added in the SecretsApp API.
    """

    # getting version through status call works
    YK_STATUS = 0x03
    status, data = helper_send_receive_ins(secretsAppRaw, YK_STATUS, le=6)
    assert len(data) == 6
    # assert data.hex()[:6] == "040b00"

    # getting serial number works
    YK_API_REQ = 0x01
    YK_P1_CMD_GET_SERIAL = 0x10
    status, data = helper_send_receive_ins(
        secretsAppRaw, YK_API_REQ, p1=YK_P1_CMD_GET_SERIAL, le=4
    )
    assert len(data) == 4

    # test HMAC calculation calls
    secretsAppRaw.reset()

    YK_P1_CMD_HMAC_1 = 0x30
    YK_P1_CMD_HMAC_2 = 0x38

    # calculation on the special-named slots does not work on factory-reset state
    for slot in [YK_P1_CMD_HMAC_2, YK_P1_CMD_HMAC_1]:
        helper_send_receive_ins(
            secretsAppRaw,
            YK_API_REQ,
            p1=slot,
            le=20,
            expected_SW="6a82",
            data_raw=helper_get_padded(b"1"),
        )

    # registration on the special-named slots works
    for i, slot in enumerate([b"HmacSlot2", b"HmacSlot1"]):
        secretsAppRaw.register(
            slot,
            secret=i.to_bytes(1, "little") * 20,
            kind=Kind.Hmac,
            pin_based_encryption=False,
        )

    for slot in [YK_P1_CMD_HMAC_2, YK_P1_CMD_HMAC_1]:
        # calculation on the hmac slot works
        status, data = helper_send_receive_ins(
            secretsAppRaw,
            YK_API_REQ,
            p1=slot,
            le=20,
            data_raw=helper_get_padded(slot.to_bytes(1, "little") * 63),
        )
        assert len(data) == 20

        # different input gives different output
        status, data = helper_send_receive_ins(
            secretsAppRaw,
            YK_API_REQ,
            p1=slot,
            le=20,
            data_raw=helper_get_padded(b"1" * 63),
        )
        status, data2 = helper_send_receive_ins(
            secretsAppRaw,
            YK_API_REQ,
            p1=slot,
            le=20,
            data_raw=helper_get_padded(b"2" * 63),
        )
        assert data != data2

        # same input gives same output
        status, data3 = helper_send_receive_ins(
            secretsAppRaw,
            YK_API_REQ,
            p1=slot,
            le=20,
            data_raw=helper_get_padded(b"2" * 63),
        )
        assert data3 == data2

    # As the last step, set the secret as the one used in the KeepassXC tests
    # $ ninja testykchallengeresponsekey && ./tests/testykchallengeresponsekey
    secret = binascii.a2b_hex(
        "1c e3 0f d7 8d 20 dc fa 40 b5 0c 18 77 9a fb 0f 02 28 8d b7".replace(" ", "")
    )
    for slot_name in [b"HmacSlot2", b"HmacSlot1"]:
        with suppress(SecretsAppException):
            secretsAppRaw.delete(slot_name)
        secretsAppRaw.register(
            slot_name,
            secret=secret,
            kind=Kind.Hmac,
        )

    with suppress(SecretsAppException):
        secretsAppRaw.delete(slot_name)

    # Do not allow to register secret with different lengths than expected 20 bytes
    for secret_len in [18, 21, 200]:
        with pytest.raises(SecretsAppException, match="IncorrectDataParameter"):
            secretsAppRaw.register(
                b"HmacSlot2",
                secret=b"x" * secret_len,
                kind=Kind.Hmac,
            )

    # ... or with Algorithm different from SHA1
    for algo in [Algorithm.Sha256]:  # Algorithm.Sha512
        with pytest.raises(SecretsAppException, match="IncorrectDataParameter"):
            secretsAppRaw.register(
                b"HmacSlot2", secret=b"x" * 20, kind=Kind.Hmac, algo=algo
            )

    # Test various challenge lengths against local calculations

    """
    Comment from the KeepassXC implementation, copied verbatim:
    /*
     * The challenge sent to the Yubikey should always be 64 bytes for
     * compatibility with all configurations.  Follow PKCS7 padding.
     *
     * There is some question whether or not 64 bytes fixed length
     * configurations even work, some docs say avoid it.
     *
     * In fact, the Yubikey always assumes the last byte (nr. 64)
     * and all bytes of the same value preceding it to be padding.
     * This does not conform fully to PKCS7, because the the actual value
     * of the padding bytes is ignored.
     */
    """

    # Prepare HmacSlot2 again
    slot_name = b"HmacSlot2"
    with suppress(SecretsAppException):
        secretsAppRaw.delete(slot_name)
    secretsAppRaw.register(
        slot_name,
        secret=secret,
        kind=Kind.Hmac,
    )

    # The length of "1" is used by KeepassXC for the purposes. "63" is the maximum.
    # "64" should not work, as the last byte is always treated as the padding byte value.
    for challenge_len in [1, 32, 63]:
        challenge = b"c" * challenge_len
        challenge_padded = helper_get_padded(challenge)
        status, response_device = helper_send_receive_ins(
            secretsAppRaw,
            YK_API_REQ,
            p1=YK_P1_CMD_HMAC_2,
            le=20,
            data_raw=challenge_padded,
        )
        response_lib = secretsAppRaw.get_response_for_secret(challenge, secret)
        assert response_lib == response_device


@pytest.mark.parametrize("touch", [True, False], ids=lambda x: "touch" if x else "")
@pytest.mark.parametrize("pws", [True, False], ids=lambda x: "pws" if x else "")
def test_list_with_properties(secretsAppResetLogin, touch, pws):
    """Test list commands with additional properties field
    @param touch: should touch_button request be set
    @param pws: should Password Safe fields be populated
    """
    length = 20
    secretsApp = secretsAppResetLogin
    login = b"login".center(length, b"=")
    password = b"password".center(length, b"=")
    metadata = b"metadata".center(length, b"=")
    name = CREDID.center(length, "=").encode()
    secretb = binascii.a2b_hex(SECRET)

    secretsApp.verify_pin_raw(PIN)
    secretsApp.register(
        name,
        secretb,
        digits=6,
        kind=Kind.Totp,
        algo=Algorithm.Sha1,
        login=login if pws else None,
        password=password if pws else None,
        metadata=metadata if pws else None,
        touch_button_required=touch,
    )

    secretsApp.verify_pin_raw(PIN)
    items_list = secretsApp.list_with_properties()
    assert len(items_list) == 1
    item = items_list[0]
    assert item.properties.touch_required == touch
    assert item.properties.pws_data_exist == pws
    assert item.properties.secret_encryption == (
        secretsAppResetLogin._metadata.get("fixture_type")
        == CredEncryptionType.PinBased
    )


def test_light_load(secretsAppRaw):
    """
    Add a couple of different Credentials' types for the manual CLI listing tests.
    When `NKPRE` env variable is set, the factory reset is not called, and Credentials
    are prefixed with this string.
    Example call:
    ```
    env NKPRE=0.12-2- make secrets-test LOG=debug TESTADD="-k light"
    ```
    """
    secretb = binascii.a2b_hex(SECRET)
    pre = environ.get("NKPRE", "")
    if not pre:
        secretsAppRaw.reset()
        secretsAppRaw.set_pin_raw(PIN)
    else:
        print(f"Using {pre} prefix for credential names")

    for encrypted in [True, False]:
        for touch in [True, False]:
            for kind in [Kind.Totp, Kind.Hotp]:
                ks = "h" if kind == Kind.Hotp else "t"
                secretsAppRaw.verify_pin_raw(PIN)
                secretsAppRaw.register(
                    f'{pre}otp:{"t" if touch else ""}:{"enc" if encrypted else ""}:{ks}'.encode(),
                    secretb,
                    digits=6,
                    kind=kind,
                    algo=Algorithm.Sha1,
                    touch_button_required=touch,
                    pin_based_encryption=encrypted,
                )

    # reverse OTP only
    secretsAppRaw.verify_pin_raw(PIN)
    secretsAppRaw.register(
        f"{pre}reverse".encode(),
        secretb,
        digits=6,
        kind=Kind.HotpReverse,
        algo=Algorithm.Sha1,
        touch_button_required=False,
    )

    # OTP + PWS test
    for pws in [True, False]:
        secretsAppRaw.register(
            f'{pre}otp:{"pws" if pws else "e"}'.encode(),
            secretb,
            digits=6,
            kind=Kind.Totp,
            algo=Algorithm.Sha1,
            login="login" if pws else None,
            password="password" if pws else None,
            metadata="metadata" if pws else None,
            touch_button_required=False,
        )

    # PWS only test
    secretsAppRaw.register(
        f"{pre}pws".encode(),
        login=b"login",
        password=b"password",
        metadata=b"metadata",
    )


@pytest.mark.parametrize(
    "cred1_encryption", [True, False], ids=lambda x: "cred1" + ("_enc" if x else "")
)
@pytest.mark.parametrize(
    "cred2_encryption", [True, False], ids=lambda x: "cred2" + ("_enc" if x else "")
)
def test_register_overwrite_attempt(secretsAppRaw, cred1_encryption, cred2_encryption):
    """
    Credential should not be overwritten with another one, regardless whether it is PIN-encrypted or not
    """
    app = secretsAppRaw
    app.reset()
    app.set_pin_raw(PIN)

    if cred1_encryption:
        app.verify_pin_raw(PIN)
    app.register(
        CREDID, SECRET, DIGITS, pin_based_encryption=cred1_encryption, kind=Kind.Hotp
    )
    if cred2_encryption:
        app.verify_pin_raw(PIN)
    with pytest.raises(SecretsAppException, match="OperationBlocked"):
        app.register(
            CREDID,
            SECRET,
            DIGITS,
            pin_based_encryption=cred2_encryption,
            kind=Kind.Hotp,
        )


def test_rename_credential(secretsAppResetLogin):
    """
    Credential should change its name. Test both PIN- and HW-encrypted credentials.
    """
    app = secretsAppResetLogin
    app.register(CREDID, SECRET, DIGITS, kind=Kind.Hotp)
    app.verify_pin_raw(PIN)
    l = app.list()
    assert len(l) == 1
    assert l[0].decode() == CREDID
    app.verify_pin_raw(PIN)
    app.rename_credential(CREDID, CREDID2)
    app.verify_pin_raw(PIN)
    # There should be only 1 credential left, with a new name
    l = app.list()
    assert len(l) == 1
    assert l[0].decode() == CREDID2

    # Old name should not be accessible
    app.verify_pin_raw(PIN)
    with pytest.raises(SecretsAppException, match="NotFound"):
        app.get_credential(CREDID)


@pytest.mark.parametrize(
    "cred1_encryption", [True, False], ids=lambda x: "cred1" + ("_enc" if x else "")
)
@pytest.mark.parametrize(
    "cred2_encryption", [True, False], ids=lambda x: "cred2" + ("_enc" if x else "")
)
def test_rename_credential_to_existing(
    secretsAppRaw, cred2_encryption, cred1_encryption
):
    """
    Credential should not change its name to an existing one, regardless if the other is PIN-encrypted or not
    """
    app = secretsAppRaw
    app.reset()
    app.set_pin_raw(PIN)

    if cred2_encryption:
        app.verify_pin_raw(PIN)
    app.register(
        CREDID2, SECRET, DIGITS, kind=Kind.Hotp, pin_based_encryption=cred2_encryption
    )
    app.verify_pin_raw(PIN)
    assert app.list()[0].decode() == CREDID2

    if cred1_encryption:
        app.verify_pin_raw(PIN)
    app.register(
        CREDID, SECRET, DIGITS, kind=Kind.Hotp, pin_based_encryption=cred1_encryption
    )
    app.verify_pin_raw(PIN)
    # Once set up, there should be 2 credentials
    assert set([CREDID.encode(), CREDID2.encode()]) == set(app.list())
    app.verify_pin_raw(PIN)
    # The rename operation should fail
    with pytest.raises(SecretsAppException, match="OperationBlocked"):
        app.rename_credential(CREDID, CREDID2)
    app.verify_pin_raw(PIN)
    # There should be still 2 credentials left
    assert set([CREDID.encode(), CREDID2.encode()]) == set(app.list())


def test_update_credential(secretsAppResetLogin):
    """
    Credential should change its properties. Test both PIN- and HW-encrypted credentials.
    """
    app = secretsAppResetLogin
    app.register(CREDID, SECRET, DIGITS, kind=Kind.Hotp)
    app.verify_pin_raw(PIN)
    l = app.list_with_properties()
    assert len(l) == 1
    assert l[0].label.decode() == CREDID
    assert not l[0].properties.touch_required
    app.verify_pin_raw(PIN)
    app.update_credential(CREDID, touch_button=True)
    # There should be only one credential left, with the same name
    app.verify_pin_raw(PIN)
    l = app.list_with_properties()
    assert len(l) == 1
    assert l[0].label.decode() == CREDID
    # This credential should be listed now as requiring a touch button for use
    assert l[0].properties.touch_required

    # Now add some PWS fields, and rename it too
    app.verify_pin_raw(PIN)
    c = app.get_credential(CREDID)
    assert not c.login
    assert not c.password
    assert not c.metadata
    app.verify_pin_raw(PIN)
    app.update_credential(
        CREDID,
        new_name=CREDID2,
        login=b"login",
        password=b"password",
        metadata=b"metadata",
    )

    # Check if PWS fields are there, and the "touch button required" flag is still present
    app.verify_pin_raw(PIN)
    c = app.get_credential(CREDID2)
    assert c.login == b"login"
    assert c.password == b"password"
    assert c.metadata == b"metadata"
    app.verify_pin_raw(PIN)
    l = app.list_with_properties()
    assert len(l) == 1
    assert l[0].properties.touch_required

    # Old name should not be accessible
    app.verify_pin_raw(PIN)
    with pytest.raises(SecretsAppException, match="NotFound"):
        app.get_credential(CREDID)

    # Try to remove the PWS data with empty strings, and rename again
    app.verify_pin_raw(PIN)
    app.update_credential(
        CREDID2, new_name=CREDID, login=b"", password=b"", metadata=b""
    )
    app.verify_pin_raw(PIN)
    c = app.get_credential(CREDID)
    assert c.login is None
    assert c.password is None
    assert c.metadata is None

    # Disallow to register a PWS credential with any 0-length strings field
    app.verify_pin_raw(PIN)
    with pytest.raises(SecretsAppException, match="IncorrectDataParameter"):
        app.register(
            CREDID2,
            SECRET,
            DIGITS,
            kind=Kind.Hotp,
            login=b"",
            password=b"",
            metadata=b"",
        )
    for i in ["login", "password", "metadata"]:
        fields = {i: b""}
        with pytest.raises(SecretsAppException, match="IncorrectDataParameter"):
            app.register(CREDID2, SECRET, DIGITS, kind=Kind.Hotp, **fields)
